#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2018, Dag Wieers (@dagwieers) <dag@wieers.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: mso_label
short_description: Manage labels
description:
- Manage labels on Cisco ACI Multi-Site.
author:
- Dag Wieers (@dagwieers)
options:
  label:
    description:
    - The name of the label.
    type: str
    aliases: [ name ]
  type:
    description:
    - The type of the label.
    type: str
    choices: [ site ]
    default: site
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment: cisco.mso.modules
'''

EXAMPLES = r'''
- name: Add a new label
  cisco.mso.mso_label:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    label: Belgium
    type: site
    state: present
  delegate_to: localhost

- name: Remove a label
  cisco.mso.mso_label:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    label: Belgium
    state: absent
  delegate_to: localhost

- name: Query a label
  cisco.mso.mso_label:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    label: Belgium
    state: query
  delegate_to: localhost
  register: query_result

- name: Query all labels
  cisco.mso.mso_label:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost
  register: query_result
'''

RETURN = r'''
'''

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        label=dict(type='str', aliases=['name']),
        type=dict(type='str', default='site', choices=['site']),
        state=dict(type='str', default='present', choices=['absent', 'present', 'query']),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ['state', 'absent', ['label']],
            ['state', 'present', ['label']],
        ],
    )

    label = module.params.get('label')
    label_type = module.params.get('type')
    state = module.params.get('state')

    mso = MSOModule(module)

    label_id = None
    path = 'labels'

    # Query for existing object(s)
    if label:
        mso.existing = mso.get_obj(path, displayName=label)
        if mso.existing:
            label_id = mso.existing.get('id')
            # If we found an existing object, continue with it
            path = 'labels/{id}'.format(id=label_id)
    else:
        mso.existing = mso.query_objs(path)

    if state == 'query':
        pass

    elif state == 'absent':
        mso.previous = mso.existing
        if mso.existing:
            if module.check_mode:
                mso.existing = {}
            else:
                mso.existing = mso.request(path, method='DELETE')

    elif state == 'present':
        mso.previous = mso.existing

        payload = dict(
            id=label_id,
            displayName=label,
            type=label_type,
        )

        mso.sanitize(payload, collate=True)

        if mso.existing:
            if mso.check_changed():
                if module.check_mode:
                    mso.existing = mso.proposed
                else:
                    mso.existing = mso.request(path, method='PUT', data=mso.sent)
        else:
            if module.check_mode:
                mso.existing = mso.proposed
            else:
                mso.existing = mso.request(path, method='POST', data=mso.sent)

    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, division, print_function

__metaclass__ = type


DOCUMENTATION = """
module: nxos_acl
extends_documentation_fragment:
- cisco.nxos.nxos
short_description: (deprecated, removed after 2022-06-01) Manages access
  list entries for ACLs.
description:
- Manages access list entries for ACLs.
version_added: 1.0.0
author:
- Jason Edelman (@jedelman8)
- Gabriele Gerbino (@GGabriele)
deprecated:
  alternative: nxos_acls
  why: Updated modules released with more functionality
  removed_at_date: '2022-06-01'
notes:
- Tested against NXOSv 7.3.(0)D1(1) on VIRL
- C(state=absent) removes the ACE if it exists.
- C(state=delete_acl) deletes the ACL if it exists.
- For idempotency, use port numbers for the src/dest port params like I(src_port1)
  and names for the well defined protocols for the I(proto) param.
- Although this module is idempotent in that if the ace as presented in the task is
  identical to the one on the switch, no changes will be made. If there is any difference,
  what is in Ansible will be pushed (configured options will be overridden).  This
  is to improve security, but at the same time remember an ACE is removed, then re-added,
  so if there is a change, the new ACE will be exactly what parameters you are sending
  to the module.
options:
  seq:
    description:
    - Sequence number of the entry (ACE).
    type: str
  name:
    description:
    - Case sensitive name of the access list (ACL).
    required: true
    type: str
  action:
    description:
    - Action of the ACE.
    choices:
    - permit
    - deny
    - remark
    type: str
  remark:
    description:
    - If action is set to remark, this is the description.
    type: str
  proto:
    description:
    - Port number or protocol (as supported by the switch).
    type: str
  src:
    description:
    - Source ip and mask using IP/MASK notation and supports keyword 'any'.
    type: str
  src_port_op:
    description:
    - Source port operands such as eq, neq, gt, lt, range.
    choices:
    - any
    - eq
    - gt
    - lt
    - neq
    - range
    type: str
  src_port1:
    description:
    - Port/protocol and also first (lower) port when using range operand.
    type: str
  src_port2:
    description:
    - Second (end) port when using range operand.
    type: str
  dest:
    description:
    - Destination ip and mask using IP/MASK notation and supports the keyword 'any'.
    type: str
  dest_port_op:
    description:
    - Destination port operands such as eq, neq, gt, lt, range.
    choices:
    - any
    - eq
    - gt
    - lt
    - neq
    - range
    type: str
  dest_port1:
    description:
    - Port/protocol and also first (lower) port when using range operand.
    type: str
  dest_port2:
    description:
    - Second (end) port when using range operand.
    type: str
  log:
    description:
    - Log matches against this entry.
    choices:
    - enable
    type: str
  urg:
    description:
    - Match on the URG bit.
    choices:
    - enable
    type: str
  ack:
    description:
    - Match on the ACK bit.
    choices:
    - enable
    type: str
  psh:
    description:
    - Match on the PSH bit.
    choices:
    - enable
    type: str
  rst:
    description:
    - Match on the RST bit.
    choices:
    - enable
    type: str
  syn:
    description:
    - Match on the SYN bit.
    choices:
    - enable
    type: str
  fin:
    description:
    - Match on the FIN bit.
    choices:
    - enable
    type: str
  established:
    description:
    - Match established connections.
    choices:
    - enable
    type: str
  fragments:
    description:
    - Check non-initial fragments.
    choices:
    - enable
    type: str
  time_range:
    description:
    - Name of time-range to apply.
    type: str
  precedence:
    description:
    - Match packets with given precedence.
    choices:
    - critical
    - flash
    - flash-override
    - immediate
    - internet
    - network
    - priority
    - routine
    type: str
  dscp:
    description:
    - Match packets with given dscp value.
    choices:
    - af11
    - af12
    - af13
    - af21
    - af22
    - af23
    - af31
    - af32
    - af33
    - af41
    - af42
    - af43
    - cs1
    - cs2
    - cs3
    - cs4
    - cs5
    - cs6
    - cs7
    - default
    - ef
    type: str
  state:
    description:
    - Specify desired state of the resource.
    default: present
    choices:
    - present
    - absent
    - delete_acl
    type: str


"""

EXAMPLES = """
# configure ACL ANSIBLE
- cisco.nxos.nxos_acl:
    name: ANSIBLE
    seq: 10
    action: permit
    proto: tcp
    src: 192.0.2.1/24
    dest: any
    state: present
"""

RETURN = """
commands:
    description: commands sent to the device
    returned: always
    type: list
    sample: ["ip access-list ANSIBLE", "10 permit tcp 192.0.2.1/24 any"]
"""
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.nxos import (
    load_config,
    run_commands,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.nxos import (
    nxos_argument_spec,
)
from ansible.module_utils.basic import AnsibleModule


def execute_show_command(command, module, check_rc=True):
    command += " | json"
    cmds = [command]
    body = run_commands(module, cmds, check_rc=check_rc)
    return body


def get_acl(module, acl_name, seq_number):
    command = "show ip access-list"
    new_acl = []
    saveme = {}
    acl_body = {}

    body = execute_show_command(command, module, check_rc=False)
    if "Structured output unsupported" in repr(body):
        # Some older versions raise 501 and return a string when no ACLs exist
        return {}, []

    if body and body[0]:
        all_acl_body = body[0]["TABLE_ip_ipv6_mac"]["ROW_ip_ipv6_mac"]
    else:
        # no access-lists configured on the device
        return {}, []

    if isinstance(all_acl_body, dict):
        # Only 1 ACL configured.
        if all_acl_body.get("acl_name") == acl_name:
            acl_body = all_acl_body
    else:
        for acl in all_acl_body:
            if acl.get("acl_name") == acl_name:
                acl_body = acl
                break

    try:
        acl_entries = acl_body["TABLE_seqno"]["ROW_seqno"]
        acl_name = acl_body.get("acl_name")
    except KeyError:  # could be raised if no ACEs are configured for an ACL
        return {}, [{"acl": "no_entries"}]

    if isinstance(acl_entries, dict):
        acl_entries = [acl_entries]

    for each in acl_entries:
        temp = {}
        options = {}
        remark = each.get("remark")

        temp["name"] = acl_name
        temp["seq"] = str(each.get("seqno"))

        if remark:
            temp["remark"] = remark
            temp["action"] = "remark"
        else:
            temp["action"] = each.get("permitdeny")
            temp["proto"] = str(
                each.get("proto", each.get("proto_str", each.get("ip")))
            )
            temp["src"] = each.get("src_any", each.get("src_ip_prefix"))
            temp["src_port_op"] = each.get("src_port_op")
            temp["src_port1"] = each.get("src_port1_num")
            temp["src_port2"] = each.get("src_port2_num")
            temp["dest"] = each.get("dest_any", each.get("dest_ip_prefix"))
            temp["dest_port_op"] = each.get("dest_port_op")
            temp["dest_port1"] = each.get("dest_port1_num")
            temp["dest_port2"] = each.get("dest_port2_num")

            options["log"] = each.get("log")
            options["urg"] = each.get("urg")
            options["ack"] = each.get("ack")
            options["psh"] = each.get("psh")
            options["rst"] = each.get("rst")
            options["syn"] = each.get("syn")
            options["fin"] = each.get("fin")
            options["established"] = each.get("established")
            options["dscp"] = each.get("dscp_str")
            options["precedence"] = each.get("precedence_str")
            options["fragments"] = each.get("fragments")
            options["time_range"] = each.get("timerange")

        keep = {}
        for key, value in temp.items():
            if value:
                keep[key] = value

        options_no_null = {}
        for key, value in options.items():
            if value is not None:
                options_no_null[key] = value
        keep["options"] = options_no_null

        if keep.get("seq") == seq_number:
            saveme = dict(keep)

        new_acl.append(keep)

    return saveme, new_acl


def _acl_operand(operand, srcp1, sprcp2):
    sub_entry = " " + operand

    if operand == "range":
        sub_entry += " " + srcp1 + " " + sprcp2
    else:
        sub_entry += " " + srcp1

    return sub_entry


def config_core_acl(proposed):
    seq = proposed.get("seq")
    action = proposed.get("action")
    remark = proposed.get("remark")
    proto = proposed.get("proto")
    src = proposed.get("src")
    src_port_op = proposed.get("src_port_op")
    src_port1 = proposed.get("src_port1")
    src_port2 = proposed.get("src_port2")

    dest = proposed.get("dest")
    dest_port_op = proposed.get("dest_port_op")
    dest_port1 = proposed.get("dest_port1")
    dest_port2 = proposed.get("dest_port2")

    ace_start_entries = [action, proto, src]
    if not remark:
        ace = seq + " " + " ".join(ace_start_entries)
        if src_port_op:
            ace += _acl_operand(src_port_op, src_port1, src_port2)
        ace += " " + dest
        if dest_port_op:
            ace += _acl_operand(dest_port_op, dest_port1, dest_port2)
    else:
        ace = seq + " remark " + remark

    return ace


def config_acl_options(options):
    ENABLE_ONLY = [
        "psh",
        "urg",
        "log",
        "ack",
        "syn",
        "established",
        "rst",
        "fin",
        "fragments",
        "log",
    ]

    OTHER = ["dscp", "precedence", "time-range"]
    # packet-length is the only option not currently supported

    if options.get("time_range"):
        options["time-range"] = options.get("time_range")
        options.pop("time_range")

    command = ""
    for option, value in options.items():
        if option in ENABLE_ONLY:
            if value == "enable":
                command += " " + option
        elif option in OTHER:
            command += " " + option + " " + value
    if command:
        command = command.strip()
        return command


def flatten_list(command_lists):
    flat_command_list = []
    for command in command_lists:
        if isinstance(command, list):
            flat_command_list.extend(command)
        else:
            flat_command_list.append(command)
    return flat_command_list


def main():
    argument_spec = dict(
        seq=dict(required=False, type="str"),
        name=dict(required=True, type="str"),
        action=dict(required=False, choices=["remark", "permit", "deny"]),
        remark=dict(required=False, type="str"),
        proto=dict(required=False, type="str"),
        src=dict(required=False, type="str"),
        src_port_op=dict(
            required=False, choices=["any", "eq", "gt", "lt", "neq", "range"]
        ),
        src_port1=dict(required=False, type="str"),
        src_port2=dict(required=False, type="str"),
        dest=dict(required=False, type="str"),
        dest_port_op=dict(
            required=False, choices=["any", "eq", "gt", "lt", "neq", "range"]
        ),
        dest_port1=dict(required=False, type="str"),
        dest_port2=dict(required=False, type="str"),
        log=dict(required=False, choices=["enable"]),
        urg=dict(required=False, choices=["enable"]),
        ack=dict(required=False, choices=["enable"]),
        psh=dict(required=False, choices=["enable"]),
        rst=dict(required=False, choices=["enable"]),
        syn=dict(required=False, choices=["enable"]),
        fragments=dict(required=False, choices=["enable"]),
        fin=dict(required=False, choices=["enable"]),
        established=dict(required=False, choices=["enable"]),
        time_range=dict(required=False),
        precedence=dict(
            required=False,
            choices=[
                "critical",
                "flash",
                "flash-override",
                "immediate",
                "internet",
                "network",
                "priority",
                "routine",
            ],
        ),
        dscp=dict(
            required=False,
            choices=[
                "af11",
                "af12",
                "af13",
                "af21",
                "af22",
                "af23",
                "af31",
                "af32",
                "af33",
                "af41",
                "af42",
                "af43",
                "cs1",
                "cs2",
                "cs3",
                "cs4",
                "cs5",
                "cs6",
                "cs7",
                "default",
                "ef",
            ],
        ),
        state=dict(
            choices=["absent", "present", "delete_acl"], default="present"
        ),
    )

    argument_spec.update(nxos_argument_spec)

    module = AnsibleModule(
        argument_spec=argument_spec, supports_check_mode=True
    )

    warnings = list()

    results = dict(changed=False, warnings=warnings)

    state = module.params["state"]
    action = module.params["action"]
    remark = module.params["remark"]
    dscp = module.params["dscp"]
    precedence = module.params["precedence"]
    seq = module.params["seq"]
    name = module.params["name"]
    seq = module.params["seq"]

    if action == "remark" and not remark:
        module.fail_json(
            msg="when state is action, remark param is also required"
        )

    REQUIRED = ["seq", "name", "action", "proto", "src", "dest"]
    ABSENT = ["name", "seq"]
    if state == "present":
        if action and remark and seq:
            pass
        else:
            for each in REQUIRED:
                if module.params[each] is None:
                    module.fail_json(
                        msg="req'd params when state is present:",
                        params=REQUIRED,
                    )
    elif state == "absent":
        for each in ABSENT:
            if module.params[each] is None:
                module.fail_json(
                    msg="require params when state is absent", params=ABSENT
                )
    elif state == "delete_acl":
        if module.params["name"] is None:
            module.fail_json(msg="param name req'd when state is delete_acl")

    if dscp and precedence:
        module.fail_json(
            msg="only one of the params dscp/precedence " "are allowed"
        )

    OPTIONS_NAMES = [
        "log",
        "urg",
        "ack",
        "psh",
        "rst",
        "syn",
        "fin",
        "established",
        "dscp",
        "precedence",
        "fragments",
        "time_range",
    ]

    CORE = [
        "seq",
        "name",
        "action",
        "proto",
        "src",
        "src_port_op",
        "src_port1",
        "src_port2",
        "dest",
        "dest_port_op",
        "dest_port1",
        "dest_port2",
        "remark",
    ]

    proposed_core = dict(
        (param, value)
        for (param, value) in module.params.items()
        if param in CORE and value is not None
    )

    proposed_options = dict(
        (param, value)
        for (param, value) in module.params.items()
        if param in OPTIONS_NAMES and value is not None
    )
    proposed = {}
    proposed.update(proposed_core)
    proposed.update(proposed_options)

    existing_options = {}

    # getting existing existing_core=dict, acl=list, seq=list
    existing_core, acl = get_acl(module, name, seq)
    if existing_core:
        existing_options = existing_core.get("options")
        existing_core.pop("options")

    commands = []
    delta_core = {}
    delta_options = {}

    if not existing_core.get("remark"):
        dcore = dict(
            set(proposed_core.items()).difference(existing_core.items())
        )
        if not dcore:
            # check the diff in the other way just in case
            dcore = dict(
                set(existing_core.items()).difference(proposed_core.items())
            )
        delta_core = dcore
        if delta_core:
            delta_options = proposed_options
        else:
            doptions = dict(
                set(proposed_options.items()).difference(
                    existing_options.items()
                )
            )
            # check the diff in the other way just in case
            if not doptions:
                doptions = dict(
                    set(existing_options.items()).difference(
                        proposed_options.items()
                    )
                )
            delta_options = doptions
    else:
        delta_core = dict(
            set(proposed_core.items()).difference(existing_core.items())
        )

    if state == "present":
        if delta_core or delta_options:
            if existing_core:  # if the ace exists already
                commands.append(["no {0}".format(seq)])
            if delta_options:
                myacl_str = config_core_acl(proposed_core)
                myacl_str += " " + config_acl_options(proposed_options)
            else:
                myacl_str = config_core_acl(proposed_core)
            command = [myacl_str]
            commands.append(command)
    elif state == "absent":
        if existing_core:
            commands.append(["no {0}".format(seq)])
    elif state == "delete_acl":
        if acl and acl[0].get("acl") != "no_entries":
            commands.append(["no ip access-list {0}".format(name)])

    cmds = []
    if commands:
        preface = []
        if state in ["present", "absent"]:
            preface = ["ip access-list {0}".format(name)]
            commands.insert(0, preface)

        cmds = flatten_list(commands)
        if module.check_mode:
            module.exit_json(changed=True, commands=cmds)
        else:
            load_config(module, cmds)
            results["changed"] = True
            if "configure" in cmds:
                cmds.pop(0)

    results["commands"] = cmds

    module.exit_json(**results)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Ansible module to manage PaloAltoNetworks Firewall
# (c) 2019, Tomi Raittinen <tomi.raittinen@gmail.com>
# (c) 2016, techbizdev <techbizdev@paloaltonetworks.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = '''
---
module: panos_commit
short_description: commit firewall's candidate configuration
description:
    - PanOS module that will commit firewall's candidate configuration on
    - the device. The new configuration will become active immediately.
author:
    - Luigi Mori (@jtschichold)
    - Ivan Bojer (@ivanbojer)
    - Tomi Raittinen (@traittinen)
requirements:
    - pan-python
deprecated:
    alternative: Use U(https://galaxy.ansible.com/PaloAltoNetworks/paloaltonetworks) instead.
    removed_in: 2.0.0  # was Ansible 2.12
    why: Consolidating code base.
options:
    ip_address:
        description:
            - IP address (or hostname) of PAN-OS device.
        required: true
    password:
        description:
            - Password for authentication. If the value is not specified in the
              task, the value of environment variable C(ANSIBLE_NET_PASSWORD)
              will be used instead.
        required: true
    username:
        description:
            - Username for authentication. If the value is not specified in the
              task, the value of environment variable C(ANSIBLE_NET_USERNAME)
              will be used instead if defined. C(admin) will be used if nothing
              above is defined.
        default: admin
    interval:
        description:
            - interval for checking commit job
        default: 0.5
    timeout:
        description:
            - timeout for commit job
    sync:
        description:
            - if commit should be synchronous
        type: bool
        default: 'yes'
    description:
        description:
            - Commit description/comment
        type: str
    commit_changes_by:
        description:
            - Commit changes made by specified admin
        type: list
    commit_vsys:
        description:
            - Commit changes for specified VSYS
        type: list
'''

EXAMPLES = '''
- name: Commit candidate config on 192.168.1.1 in sync mode
  community.network.panos_commit:
    ip_address: "192.168.1.1"
    username: "admin"
    password: "admin"
'''

RETURN = '''
panos_commit:
    description: Information about commit job.
    returned: always
    type: complex
    contains:
        job_id:
            description: Palo Alto job ID for the commit operation. Only returned if commit job is launched on device.
            returned: always
            type: str
            sample: "139"
        status_code:
            description: Palo Alto API status code. Null if commit is successful.
            returned: always
            type: str
            sample: 19
        status_detail:
            description: Palo Alto API detailed status message.
            returned: always
            type: str
            sample: Configuration committed successfully
        status_text:
            description: Palo Alto API status text.
            returned: always
            type: str
            sample: success
'''

from ansible.module_utils.basic import AnsibleModule, env_fallback
import xml.etree.ElementTree as etree

try:
    import pan.xapi
    HAS_LIB = True
except ImportError:
    HAS_LIB = False


def main():
    argument_spec = dict(
        ip_address=dict(required=True, type='str'),
        password=dict(fallback=(env_fallback, ['ANSIBLE_NET_PASSWORD']), no_log=True),
        username=dict(fallback=(env_fallback, ['ANSIBLE_NET_USERNAME']), default="admin"),
        interval=dict(default=0.5),
        timeout=dict(),
        sync=dict(type='bool', default=True),
        description=dict(type='str'),
        commit_changes_by=dict(type='list'),
        commit_vsys=dict(type='list')
    )
    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=False)

    if not HAS_LIB:
        module.fail_json(msg='pan-python is required for this module')

    ip_address = module.params["ip_address"]
    if not ip_address:
        module.fail_json(msg="ip_address should be specified")

    password = module.params["password"]
    if not password:
        module.fail_json(msg="password is required")

    username = module.params['username']
    if not username:
        module.fail_json(msg="username is required")

    interval = module.params['interval']
    timeout = module.params['timeout']
    sync = module.params['sync']

    xapi = pan.xapi.PanXapi(
        hostname=ip_address,
        api_username=username,
        api_password=password
    )

    cmd = "<commit>"

    description = module.params["description"]
    if description:
        cmd += "<description>" + description + "</description>"

    commit_changes_by = module.params["commit_changes_by"]
    commit_vsys = module.params["commit_vsys"]

    if commit_changes_by or commit_vsys:

        cmd += "<partial>"

        if commit_changes_by:
            cmd += "<admin>"
            for admin in commit_changes_by:
                cmd += "<member>" + admin + "</member>"
            cmd += "</admin>"

        if commit_vsys:
            cmd += "<vsys>"
            for vsys in commit_vsys:
                cmd += "<member>" + vsys + "</member>"
            cmd += "</vsys>"

        cmd += "</partial><force></force>"

    cmd += "</commit>"

    xapi.commit(
        cmd=cmd,
        sync=sync,
        interval=interval,
        timeout=timeout
    )

    try:
        result = xapi.xml_root().encode('utf-8')
        root = etree.fromstring(result)
        job_id = root.find('./result/job/id').text
    except AttributeError:
        job_id = None

    panos_commit_details = dict(
        status_text=xapi.status,
        status_code=xapi.status_code,
        status_detail=xapi.status_detail,
        job_id=job_id
    )

    if "Commit failed" in xapi.status_detail:
        module.fail_json(msg=xapi.status_detail, panos_commit=panos_commit_details)

    if job_id:
        module.exit_json(changed=True, msg="Commit successful.", panos_commit=panos_commit_details)
    else:
        module.exit_json(changed=False, msg="No changes to commit.", panos_commit=panos_commit_details)


if __name__ == '__main__':
    main()



# Copyright: (c) 2018, Aaron Haaf <aabonh@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import datetime
import hashlib
import hmac
import operator

try:
    from boto3 import session
except ImportError:
    pass

from ansible.module_utils.six.moves.urllib.parse import urlencode
from ansible.module_utils.urls import open_url

from .ec2 import HAS_BOTO3
from .ec2 import get_aws_connection_info


def hexdigest(s):
    """
    Returns the sha256 hexdigest of a string after encoding.
    """

    return hashlib.sha256(s.encode("utf-8")).hexdigest()


def format_querystring(params=None):
    """
    Returns properly url-encoded query string from the provided params dict.

    It's specially sorted for cannonical requests
    """

    if not params:
        return ""

    # Query string values must be URL-encoded (space=%20). The parameters must be sorted by name.
    return urlencode(sorted(params.items(), operator.itemgetter(0)))


# Key derivation functions. See:
# http://docs.aws.amazon.com/general/latest/gr/signature-v4-examples.html#signature-v4-examples-python
def sign(key, msg):
    '''
    Return digest for key applied to msg
    '''

    return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()


def get_signature_key(key, dateStamp, regionName, serviceName):
    '''
    Returns signature key for AWS resource
    '''

    kDate = sign(("AWS4" + key).encode("utf-8"), dateStamp)
    kRegion = sign(kDate, regionName)
    kService = sign(kRegion, serviceName)
    kSigning = sign(kService, "aws4_request")
    return kSigning


def get_aws_credentials_object(module):
    '''
    Returns aws_access_key_id, aws_secret_access_key, session_token for a module.
    '''

    if not HAS_BOTO3:
        module.fail_json("get_aws_credentials_object requires boto3")

    dummy, dummy, boto_params = get_aws_connection_info(module, boto3=True)
    s = session.Session(**boto_params)

    return s.get_credentials()


# Reference: https://docs.aws.amazon.com/general/latest/gr/sigv4-signed-request-examples.html
def signed_request(
        module=None,
        method="GET", service=None, host=None, uri=None,
        query=None, body="", headers=None,
        session_in_header=True, session_in_query=False
):
    """Generate a SigV4 request to an AWS resource for a module

    This is used if you wish to authenticate with AWS credentials to a secure endpoint like an elastisearch domain.

    Returns :class:`HTTPResponse` object.

    Example:
        result = signed_request(
            module=this,
            service="es",
            host="search-recipes1-xxxxxxxxx.us-west-2.es.amazonaws.com",
        )

    :kwarg host: endpoint to talk to
    :kwarg service: AWS id of service (like `ec2` or `es`)
    :kwarg module: An AnsibleAWSModule to gather connection info from

    :kwarg body: (optional) Payload to send
    :kwarg method: (optional) HTTP verb to use
    :kwarg query: (optional) dict of query params to handle
    :kwarg uri: (optional) Resource path without query parameters

    :kwarg session_in_header: (optional) Add the session token to the headers
    :kwarg session_in_query: (optional) Add the session token to the query parameters

    :returns: HTTPResponse
    """

    if not HAS_BOTO3:
        module.fail_json("A sigv4 signed_request requires boto3")

    # "Constants"

    t = datetime.datetime.utcnow()
    amz_date = t.strftime("%Y%m%dT%H%M%SZ")
    datestamp = t.strftime("%Y%m%d")  # Date w/o time, used in credential scope
    algorithm = "AWS4-HMAC-SHA256"

    # AWS stuff

    region, dummy, dummy = get_aws_connection_info(module, boto3=True)
    credentials = get_aws_credentials_object(module)
    access_key = credentials.access_key
    secret_key = credentials.secret_key
    session_token = credentials.token

    if not access_key:
        module.fail_json(msg="aws_access_key_id is missing")
    if not secret_key:
        module.fail_json(msg="aws_secret_access_key is missing")

    credential_scope = "/".join([datestamp, region, service, "aws4_request"])

    # Argument Defaults

    uri = uri or "/"
    query_string = format_querystring(query) if query else ""

    headers = headers or dict()
    query = query or dict()

    headers.update({
        "host": host,
        "x-amz-date": amz_date,
    })

    # Handle adding of session_token if present
    if session_token:
        if session_in_header:
            headers["X-Amz-Security-Token"] = session_token
        if session_in_query:
            query["X-Amz-Security-Token"] = session_token

    if method == "GET":
        body = ""

    # Derived data

    body_hash = hexdigest(body)
    signed_headers = ";".join(sorted(headers.keys()))

    # Setup Cannonical request to generate auth token

    cannonical_headers = "\n".join([
        key.lower().strip() + ":" + value for key, value in headers.items()
    ]) + "\n"  # Note additional trailing newline

    cannonical_request = "\n".join([
        method,
        uri,
        query_string,
        cannonical_headers,
        signed_headers,
        body_hash,
    ])

    string_to_sign = "\n".join([algorithm, amz_date, credential_scope, hexdigest(cannonical_request)])

    # Sign the Cannonical request

    signing_key = get_signature_key(secret_key, datestamp, region, service)
    signature = hmac.new(signing_key, string_to_sign.encode("utf-8"), hashlib.sha256).hexdigest()

    # Make auth header with that info

    authorization_header = "{0} Credential={1}/{2}, SignedHeaders={3}, Signature={4}".format(
        algorithm, access_key, credential_scope, signed_headers, signature
    )

    # PERFORM THE REQUEST!

    url = "https://" + host + uri

    if query_string != "":
        url = url + "?" + query_string

    final_headers = {
        "x-amz-date": amz_date,
        "Authorization": authorization_header,
    }

    final_headers.update(headers)

    return open_url(url, method=method, data=body, headers=final_headers)



#!/usr/bin/python
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type


DOCUMENTATION = '''
---
module: ce_static_route_bfd
version_added: '0.2.0'
short_description: Manages static route configuration on HUAWEI CloudEngine switches.
description:
    - Manages the static routes on HUAWEI CloudEngine switches.
author: xuxiaowei0512 (@CloudEngine-Ansible)
notes:
  - This module requires the netconf system service be enabled on the remote device being managed.
  - Recommended connection is C(netconf).
  - This module also works with C(local) connections for legacy playbooks.
  - If no vrf is supplied, vrf is set to default.
  - If I(state=absent), the route configuration will be removed, regardless of the non-required parameters.
options:
  prefix:
    description:
      - Destination ip address of static route.
    required: true
    type: str
  mask:
   description:
     - Destination ip mask of static route.
   type: str
  aftype:
    description:
      - Destination ip address family type of static route.
    required: true
    type: str
    choices: ['v4','v6']
  next_hop:
    description:
      - Next hop address of static route.
    type: str
  nhp_interface:
    description:
      - Next hop interface full name of static route.
    type: str
  vrf:
    description:
      - VPN instance of destination ip address.
    type: str
  destvrf:
    description:
      - VPN instance of next hop ip address.
    type: str
  tag:
    description:
      - Route tag value (numeric).
    type: int
  description:
    description:
      - Name of the route. Used with the name parameter on the CLI.
    type: str
  pref:
    description:
      - Preference or administrative difference of route (range 1-255).
    type: int
  function_flag:
    description:
      - Used to distinguish between command line functions.
    required: true
    choices: ['globalBFD','singleBFD','dynamicBFD','staticBFD']
    type: str
  min_tx_interval:
    description:
      - Set the minimum BFD session sending interval (range 50-1000).
    type: int
  min_rx_interval:
    description:
      - Set the minimum BFD receive interval (range 50-1000).
    type: int
  detect_multiplier:
    description:
      - Configure the BFD multiplier (range 3-50).
    type: int
  bfd_session_name:
    description:
      - bfd name (range 1-15).
    type: str
  commands:
    description:
      - Incoming command line is used to send sys,undo ip route-static default-bfd,commit.
    type: list
  state:
    description:
      - Specify desired state of the resource.
    required: false
    choices: ['present','absent']
    type: str
    default: present
'''

EXAMPLES = '''
  #ip route-static bfd interface-type interface-number nexthop-address [ local-address address ]
  #[ min-rx-interval min-rx-interval | min-tx-interval min-tx-interval | detect-multiplier multiplier ]
  - name: Config an ip route-static bfd 10GE1/0/1 3.3.3.3 min-rx-interval 50 min-tx-interval 50 detect-multiplier 5
    community.network.ce_static_route_bfd:
      function_flag: 'singleBFD'
      nhp_interface: 10GE1/0/1
      next_hop: 3.3.3.3
      min_tx_interval: 50
      min_rx_interval: 50
      detect_multiplier: 5
      aftype: v4
      state: present

  #undo ip route-static bfd [ interface-type interface-number | vpn-instance vpn-instance-name ] nexthop-address
  - name: Undo ip route-static bfd 10GE1/0/1 3.3.3.4
    community.network.ce_static_route_bfd:
      function_flag: 'singleBFD'
      nhp_interface: 10GE1/0/1
      next_hop: 3.3.3.4
      aftype: v4
      state: absent

  #ip route-static default-bfd { min-rx-interval {min-rx-interval} | min-tx-interval {min-tx-interval} | detect-multiplier {multiplier}}
  - name: Config an ip route-static default-bfd min-rx-interval 50 min-tx-interval 50 detect-multiplier 6
    community.network.ce_static_route_bfd:
      function_flag: 'globalBFD'
      min_tx_interval: 50
      min_rx_interval: 50
      detect_multiplier: 6
      aftype: v4
      state: present

  - name: Undo ip route-static default-bfd
    community.network.ce_static_route_bfd:
      function_flag: 'globalBFD'
      aftype: v4
      state: absent
      commands: 'sys,undo ip route-static default-bfd,commit'

  - name: Config an ipv4 static route 2.2.2.0/24 2.2.2.1 preference 1 tag 2 description test for staticBFD
    community.network.ce_static_route_bfd:
      function_flag: 'staticBFD'
      prefix: 2.2.2.2
      mask: 24
      next_hop: 2.2.2.1
      tag: 2
      description: test
      pref: 1
      aftype: v4
      bfd_session_name: btoa
      state: present
'''
RETURN = '''
proposed:
    description: k/v pairs of parameters passed into module
    returned: always
    type: dict
    sample: {"function_flag": "staticBFD", "next_hop": "3.3.3.3", "pref": "100",
            "prefix": "192.168.20.642", "mask": "24", "description": "testing",
            "vrf": "_public_", "bfd_session_name": "btoa"}
existing:
    description: k/v pairs of existing switchport
    returned: always
    type: dict
    sample: {"function_flag": "", "next_hop": "", "pref": "101",
            "prefix": "192.168.20.0", "mask": "24", "description": "testing",
            "tag" : "null", "bfd_session_name": "btoa"}
end_state:
    description: k/v pairs of switchport after module execution
    returned: always
    type: dict
    sample: {"function_flag": "staticBFD", "next_hop": "3.3.3.3", "pref": "100",
            "prefix": "192.168.20.0", "mask": "24", "description": "testing",
            "tag" : "null", "bfd_session_name": "btoa"}
updates:
    description: command list sent to the device
    returned: always
    type: list
    sample: ["ip route-static 192.168.20.0 255.255.255.0 3.3.3.3 preference 100 description testing"]
changed:
    description: check to see if a change was made on the device
    returned: always
    type: bool
    sample: true
'''

from xml.etree import ElementTree
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.six import string_types
from ansible_collections.community.network.plugins.module_utils.network.cloudengine.ce import get_nc_config, set_nc_config

CE_NC_GET_STATIC_ROUTE_BFD_SESSIONNAME = """
<filter type="subtree">
  <staticrt xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
    <staticrtbase>
      <srRoutes>
        <srRoute>
          <vrfName></vrfName>
          <afType></afType>
          <topologyName></topologyName>
          <prefix></prefix>
          <maskLength></maskLength>
          <ifName></ifName>
          <destVrfName></destVrfName>
          <nexthop></nexthop>
          <preference></preference>
          <tag></tag>
          <sessionName></sessionName>
          <description></description>
        </srRoute>
      </srRoutes>
    </staticrtbase>
  </staticrt>
</filter>
"""
# bfd enable
CE_NC_GET_STATIC_ROUTE_BFD_ENABLE = """
<filter type="subtree">
      <staticrt xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
        <staticrtbase>
          <srRoutes>
            <srRoute>
              <vrfName></vrfName>
              <afType></afType>
              <topologyName></topologyName>
              <prefix></prefix>
              <maskLength></maskLength>
              <ifName></ifName>
              <destVrfName></destVrfName>
              <nexthop></nexthop>
              <preference></preference>
              <tag></tag>
              <bfdEnable></bfdEnable>
              <description></description>
            </srRoute>
          </srRoutes>
        </staticrtbase>
      </staticrt>
    </filter>
"""

CE_NC_GET_STATIC_ROUTE_BFD_ABSENT = """
<filter type="subtree">
  <staticrt xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
    <staticrtbase>
      <srBfdParas>
        <srBfdPara>
          <afType>%s</afType>
          <ifName>%s</ifName>
          <destVrfName>%s</destVrfName>
          <nexthop>%s</nexthop>
        </srBfdPara>
      </srBfdParas>
    </staticrtbase>
  </staticrt>
</filter>
"""

CE_NC_GET_STATIC_ROUTE_BFD = """
<filter type="subtree">
  <staticrt xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
    <staticrtbase>
      <srBfdParas>
        <srBfdPara>
          <afType>%s</afType>
          <ifName>%s</ifName>
          <destVrfName>%s</destVrfName>
          <nexthop>%s</nexthop>
          <localAddress></localAddress>
          <minTxInterval></minTxInterval>
          <minRxInterval></minRxInterval>
          <multiplier></multiplier>
        </srBfdPara>
      </srBfdParas>
    </staticrtbase>
  </staticrt>
</filter>
"""
CE_NC_GET_STATIC_ROUTE_IPV4_GLOBAL_BFD = """
<filter type="subtree">
  <staticrt xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
    <staticrtbase>
      <srIPv4StaticSite>
        <minTxInterval></minTxInterval>
        <minRxInterval></minRxInterval>
        <multiplier></multiplier>
      </srIPv4StaticSite>
    </staticrtbase>
  </staticrt>
</filter>
"""
CE_NC_GET_STATIC_ROUTE_ABSENT = """
<filter type="subtree">
      <staticrt xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
        <staticrtbase>
          <srRoutes>
            <srRoute>
              <vrfName></vrfName>
              <afType></afType>
              <topologyName></topologyName>
              <prefix></prefix>
              <maskLength></maskLength>
              <ifName></ifName>
              <destVrfName></destVrfName>
              <nexthop></nexthop>
            </srRoute>
          </srRoutes>
        </staticrtbase>
      </staticrt>
    </filter>
"""

CE_NC_DELETE_STATIC_ROUTE_SINGLEBFD = """
<staticrt xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
        <staticrtbase>
          <srBfdParas>
            <srBfdPara operation="delete">
              <afType>%s</afType>
              <ifName>%s</ifName>
              <destVrfName>%s</destVrfName>
              <nexthop>%s</nexthop>
            </srBfdPara>
          </srBfdParas>
        </staticrtbase>
      </staticrt>
"""
CE_NC_SET_STATIC_ROUTE_SINGLEBFD = """
<staticrt xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
        <staticrtbase>
          <srBfdParas>
            <srBfdPara operation="merge">
              <afType>%s</afType>
              <ifName>%s</ifName>
              <destVrfName>%s</destVrfName>
              <nexthop>%s</nexthop>%s%s%s%s
            </srBfdPara>
          </srBfdParas>
        </staticrtbase>
      </staticrt>

"""
CE_NC_SET_STATIC_ROUTE_SINGLEBFD_LOCALADRESS = """
<localAddress>%s</localAddress>
"""
CE_NC_SET_IPV4_STATIC_ROUTE_BFDCOMMON_MINTX = """
<minTxInterval>%s</minTxInterval>
"""
CE_NC_SET_IPV4_STATIC_ROUTE_BFDCOMMON_MINRX = """
<minRxInterval>%s</minRxInterval>
"""
CE_NC_SET_IPV4_STATIC_ROUTE_BFDCOMMON_MUL = """
<multiplier>%s</multiplier>
"""
CE_NC_SET_IPV4_STATIC_ROUTE_GLOBALBFD = """
<staticrt xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
    <staticrtbase>
      <srIPv4StaticSite operation="merge">
      %s%s%s
      </srIPv4StaticSite>
    </staticrtbase>
</staticrt>
"""

CE_NC_SET_STATIC_ROUTE = """
<staticrt xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
        <staticrtbase>
          <srRoutes>
            <srRoute operation="merge">
              <vrfName>%s</vrfName>
              <afType>%s</afType>
              <topologyName>base</topologyName>
              <prefix>%s</prefix>
              <maskLength>%s</maskLength>
              <ifName>%s</ifName>
              <destVrfName>%s</destVrfName>
              <nexthop>%s</nexthop>%s%s%s%s
            </srRoute>
          </srRoutes>
        </staticrtbase>
      </staticrt>
"""
CE_NC_SET_DESCRIPTION = """
<description>%s</description>
"""

CE_NC_SET_PREFERENCE = """
<preference>%s</preference>
"""

CE_NC_SET_TAG = """
<tag>%s</tag>
"""
CE_NC_SET_BFDSESSIONNAME = """
<sessionName>%s</sessionName>
"""
CE_NC_SET_BFDENABLE = """
<bfdEnable>true</bfdEnable>
"""
CE_NC_DELETE_STATIC_ROUTE = """
<staticrt xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
        <staticrtbase>
          <srRoutes>
            <srRoute operation="delete">
              <vrfName>%s</vrfName>
              <afType>%s</afType>
              <topologyName>base</topologyName>
              <prefix>%s</prefix>
              <maskLength>%s</maskLength>
              <ifName>%s</ifName>
              <destVrfName>%s</destVrfName>
              <nexthop>%s</nexthop>
            </srRoute>
          </srRoutes>
        </staticrtbase>
      </staticrt>
"""


def build_config_xml(xmlstr):
    """build config xml"""

    return '<config> ' + xmlstr + ' </config>'


def is_valid_v4addr(addr):
    """check if ipv4 addr is valid"""
    if addr.find('.') != -1:
        addr_list = addr.split('.')
        if len(addr_list) != 4:
            return False
        for each_num in addr_list:

            if not each_num.isdigit():
                return False
            if int(each_num) > 255:
                return False
        return True
    return False


def is_valid_v6addr(addr):
    """check if ipv6 addr is valid"""
    if addr.find(':') != -1:
        addr_list = addr.split(':')
        if len(addr_list) > 6:
            return False
        if addr_list[1] == "":
            return False
        return True
    return False


def is_valid_tag(tag):
    """check if the tag is valid"""

    if int(tag) < 1 or int(tag) > 4294967295:
        return False
    return True


def is_valid_bdf_interval(interval):
    """check if the min_tx_interva,min-rx-interval is valid"""

    if interval < 50 or interval > 1000:
        return False
    return True


def is_valid_bdf_multiplier(multiplier):
    """check if the detect_multiplier is valid"""

    if multiplier < 3 or multiplier > 50:
        return False
    return True


def is_valid_bdf_session_name(session_name):
    """check if the bfd_session_name is valid"""
    if session_name.find(' ') != -1:
        return False
    if len(session_name) < 1 or len(session_name) > 15:
        return False
    return True


def is_valid_preference(pref):
    """check if the preference is valid"""

    if int(pref) > 0 and int(pref) < 256:
        return True
    return False


def is_valid_description(description):
    """check if the description is valid"""
    if description.find('?') != -1:
        return False
    if len(description) < 1 or len(description) > 255:
        return False
    return True


def compare_command(commands):
    """check if the commands is valid"""
    if len(commands) < 3:
        return True
    if commands[0] != 'sys' or commands[1] != 'undo ip route-static default-bfd' \
            or commands[2] != 'commit':
        return True


def get_to_lines(stdout):
    """data conversion"""
    lines = list()
    for item in stdout:
        if isinstance(item, string_types):
            item = str(item).split('\n')
        lines.append(item)
    return lines


def get_change_state(oldvalue, newvalue, change):
    """get change state"""
    if newvalue is not None:
        if oldvalue != str(newvalue):
            change = True
    else:
        if oldvalue != newvalue:
            change = True
    return change


def get_xml(xml, value):
    """operate xml"""
    if value is None:
        value = ''
    else:
        value = value
    tempxml = xml % value
    return tempxml


class StaticRouteBFD(object):
    """static route module"""

    def __init__(self, argument_spec):
        self.spec = argument_spec
        self.module = None
        self._initmodule_()

        # static route info
        self.function_flag = self.module.params['function_flag']
        self.aftype = self.module.params['aftype']
        self.state = self.module.params['state']
        if self.aftype == "v4":
            self.version = "ipv4unicast"
        else:
            self.version = "ipv6unicast"
        if self.function_flag != 'globalBFD':
            self.nhp_interface = self.module.params['nhp_interface']
            if self.nhp_interface is None:
                self.nhp_interface = "Invalid0"

            self.destvrf = self.module.params['destvrf']
            if self.destvrf is None:
                self.destvrf = "_public_"

            self.next_hop = self.module.params['next_hop']
            self.prefix = self.module.params['prefix']

        if self.function_flag != 'globalBFD' and self.function_flag != 'singleBFD':
            self.mask = self.module.params['mask']
            self.tag = self.module.params['tag']
            self.description = self.module.params['description']
            self.pref = self.module.params['pref']
            if self.pref is None:
                self.pref = 60
            # vpn instance info
            self.vrf = self.module.params['vrf']
            if self.vrf is None:
                self.vrf = "_public_"
            # bfd session name
            self.bfd_session_name = self.module.params['bfd_session_name']

        if self.function_flag == 'globalBFD' or self.function_flag == 'singleBFD':
            self.min_tx_interval = self.module.params['min_tx_interval']
            self.min_rx_interval = self.module.params['min_rx_interval']
            self.detect_multiplier = self.module.params['detect_multiplier']
        if self.function_flag == 'globalBFD' and self.state == 'absent':
            self.commands = self.module.params['commands']
        # state
        self.changed = False
        self.updates_cmd = list()
        self.results = dict()
        self.proposed = dict()
        self.existing = dict()
        self.end_state = dict()

        self.static_routes_info = dict()

    def _initmodule_(self):
        """init module"""

        self.module = AnsibleModule(
            argument_spec=self.spec, supports_check_mode=False)

    def _checkresponse_(self, xml_str, xml_name):
        """check if response message is already succeed."""

        if "<ok/>" not in xml_str:
            self.module.fail_json(msg='Error: %s failed.' % xml_name)

    def _convertlentomask_(self, masklen):
        """convert mask length to ip address mask, i.e. 24 to 255.255.255.0"""

        mask_int = ["0"] * 4
        length = int(masklen)

        if length > 32:
            self.module.fail_json(msg='IPv4 ipaddress mask length is invalid')
        if length < 8:
            mask_int[0] = str(int((0xFF << (8 - length % 8)) & 0xFF))
        if length >= 8:
            mask_int[0] = '255'
            mask_int[1] = str(int((0xFF << (16 - (length % 16))) & 0xFF))
        if length >= 16:
            mask_int[1] = '255'
            mask_int[2] = str(int((0xFF << (24 - (length % 24))) & 0xFF))
        if length >= 24:
            mask_int[2] = '255'
            mask_int[3] = str(int((0xFF << (32 - (length % 32))) & 0xFF))
        if length == 32:
            mask_int[3] = '255'

        return '.'.join(mask_int)

    def _convertipprefix_(self):
        """convert prefix to real value i.e. 2.2.2.2/24 to 2.2.2.0/24"""
        if self.function_flag == 'singleBFD':
            if self.aftype == "v4":
                if self.prefix.find('.') == -1:
                    return False
                addr_list = self.prefix.split('.')
                length = len(addr_list)
                if length > 4:
                    return False
                for each_num in addr_list:
                    if not each_num.isdigit():
                        return False
                    if int(each_num) > 255:
                        return False
                return True
            else:
                if self.prefix.find(':') == -1:
                    return False
        else:
            if self.aftype == "v4":
                if self.prefix.find('.') == -1:
                    return False
                if self.mask == '32':
                    self.prefix = self.prefix
                    return True
                if self.mask == '0':
                    self.prefix = '0.0.0.0'
                    return True
                addr_list = self.prefix.split('.')
                length = len(addr_list)
                if length > 4:
                    return False
                for each_num in addr_list:
                    if not each_num.isdigit():
                        return False
                    if int(each_num) > 255:
                        return False
                byte_len = 8
                ip_len = int(self.mask) // byte_len
                ip_bit = int(self.mask) % byte_len
            else:
                if self.prefix.find(':') == -1:
                    return False
                if self.mask == '128':
                    self.prefix = self.prefix
                    return True
                if self.mask == '0':
                    self.prefix = '::'
                    return True
                addr_list = self.prefix.split(':')
                length = len(addr_list)
                if length > 6:
                    return False
                byte_len = 16
                ip_len = int(self.mask) // byte_len
                ip_bit = int(self.mask) % byte_len

            if self.aftype == "v4":
                for i in range(ip_len + 1, length):
                    addr_list[i] = 0
            else:
                for i in range(length - ip_len, length):
                    addr_list[i] = 0
            for j in range(0, byte_len - ip_bit):
                if self.aftype == "v4":
                    addr_list[ip_len] = int(addr_list[ip_len]) & (0 << j)
                else:
                    if addr_list[length - ip_len - 1] == "":
                        continue
                    addr_list[length - ip_len -
                              1] = '0x%s' % addr_list[length - ip_len - 1]
                    addr_list[length - ip_len -
                              1] = int(addr_list[length - ip_len - 1], 16) & (0 << j)

            if self.aftype == "v4":
                self.prefix = '%s.%s.%s.%s' % (addr_list[0], addr_list[1], addr_list[2], addr_list[3])
                return True
            if self.aftype == "v6":
                ipv6_addr_str = ""
                for num in range(0, length - ip_len):
                    ipv6_addr_str += '%s:' % addr_list[num]
                self.prefix = ipv6_addr_str

                return True

    def set_update_cmd_globalbfd(self):
        """set globalBFD update command"""
        if not self.changed:
            return
        if self.state == "present":
            self.updates_cmd.append('ip route-static default-bfd')
            if self.min_tx_interval:
                self.updates_cmd.append(' min-rx-interval %s' % (self.min_tx_interval))
            if self.min_rx_interval:
                self.updates_cmd.append(' min-tx-interval %s' % (self.min_rx_interval))
            if self.detect_multiplier:
                self.updates_cmd.append(' detect-multiplier %s' % (self.detect_multiplier))
        else:
            self.updates_cmd.append('undo ip route-static default-bfd')

    def set_update_cmd_singlebfd(self):
        """set singleBFD update command"""
        if not self.changed:
            return
        if self.next_hop is None:
            next_hop = ''
        else:
            next_hop = self.next_hop

        if self.destvrf == "_public_":
            destvrf = ''
        else:
            destvrf = self.destvrf

        if self.nhp_interface == "Invalid0":
            nhp_interface = ''
        else:
            nhp_interface = self.nhp_interface
        if self.prefix == "0.0.0.0":
            prefix = ''
        else:
            prefix = self.prefix
        if self.state == "present":
            if nhp_interface:
                self.updates_cmd.append('ip route-static bfd %s %s' % (nhp_interface, next_hop))
            elif destvrf:
                self.updates_cmd.append('ip route-static bfd vpn-instance %s %s' % (destvrf, next_hop))
            else:
                self.updates_cmd.append('ip route-static bfd %s' % (next_hop))
            if prefix:
                self.updates_cmd.append(' local-address %s' % (self.prefix))
            if self.min_tx_interval:
                self.updates_cmd.append(' min-rx-interval %s' % (self.min_tx_interval))
            if self.min_rx_interval:
                self.updates_cmd.append(' min-tx-interval %s' % (self.min_rx_interval))
            if self.detect_multiplier:
                self.updates_cmd.append(' detect-multiplier %s' % (self.detect_multiplier))
        else:
            if nhp_interface:
                self.updates_cmd.append('undo ip route-static bfd %s %s' % (nhp_interface, next_hop))
            elif destvrf:
                self.updates_cmd.append('undo ip route-static bfd vpn-instance %s %s' % (destvrf, next_hop))
            else:
                self.updates_cmd.append('undo ip route-static bfd %s' % (next_hop))

    def set_update_cmd(self):
        """set update command"""
        if not self.changed:
            return

        if self.aftype == "v4":
            maskstr = self._convertlentomask_(self.mask)
        else:
            maskstr = self.mask
        static_bfd_flag = True
        if self.bfd_session_name:
            static_bfd_flag = False
        if self.next_hop is None:
            next_hop = ''
        else:
            next_hop = self.next_hop
        if self.vrf == "_public_":
            vrf = ''
        else:
            vrf = self.vrf
        if self.destvrf == "_public_":
            destvrf = ''
        else:
            destvrf = self.destvrf
        if self.nhp_interface == "Invalid0":
            nhp_interface = ''
        else:
            nhp_interface = self.nhp_interface
        if self.state == "present":
            if self.vrf != "_public_":
                if self.destvrf != "_public_":
                    self.updates_cmd.append('ip route-static vpn-instance %s %s %s vpn-instance %s %s'
                                            % (vrf, self.prefix, maskstr, destvrf, next_hop))
                else:
                    self.updates_cmd.append('ip route-static vpn-instance %s %s %s %s %s'
                                            % (vrf, self.prefix, maskstr, nhp_interface, next_hop))
            elif self.destvrf != "_public_":
                self.updates_cmd.append('ip route-static %s %s vpn-instance %s %s'
                                        % (self.prefix, maskstr, self.destvrf, next_hop))
            else:
                self.updates_cmd.append('ip route-static %s %s %s %s'
                                        % (self.prefix, maskstr, nhp_interface, next_hop))
            if self.pref != 60:
                self.updates_cmd.append(' preference %s' % (self.pref))
            if self.tag:
                self.updates_cmd.append(' tag %s' % (self.tag))
            if not static_bfd_flag:
                self.updates_cmd.append(' track bfd-session %s' % (self.bfd_session_name))
            else:
                self.updates_cmd.append(' bfd enable')
            if self.description:
                self.updates_cmd.append(' description %s' % (self.description))

        if self.state == "absent":
            if self.vrf != "_public_":
                if self.destvrf != "_public_":
                    self.updates_cmd.append('undo ip route-static vpn-instance %s %s %s vpn-instance %s %s'
                                            % (vrf, self.prefix, maskstr, destvrf, next_hop))
                else:
                    self.updates_cmd.append('undo ip route-static vpn-instance %s %s %s %s %s'
                                            % (vrf, self.prefix, maskstr, nhp_interface, next_hop))
            elif self.destvrf != "_public_":
                self.updates_cmd.append('undo ip route-static %s %s vpn-instance %s %s'
                                        % (self.prefix, maskstr, self.destvrf, self.next_hop))
            else:
                self.updates_cmd.append('undo ip route-static %s %s %s %s'
                                        % (self.prefix, maskstr, nhp_interface, next_hop))

    def operate_static_route_globalbfd(self):
        """set globalbfd update command"""
        min_tx_interval = self.min_tx_interval
        min_rx_interval = self.min_rx_interval
        multiplier = self.detect_multiplier
        min_tx_interval_xml = """\n"""
        min_rx_interval_xml = """\n"""
        multiplier_xml = """\n"""
        if self.state == "present":
            if min_tx_interval is not None:
                min_tx_interval_xml = CE_NC_SET_IPV4_STATIC_ROUTE_BFDCOMMON_MINTX % min_tx_interval
            if min_rx_interval is not None:
                min_rx_interval_xml = CE_NC_SET_IPV4_STATIC_ROUTE_BFDCOMMON_MINRX % min_rx_interval
            if multiplier is not None:
                multiplier_xml = CE_NC_SET_IPV4_STATIC_ROUTE_BFDCOMMON_MUL % multiplier

            configxmlstr = CE_NC_SET_IPV4_STATIC_ROUTE_GLOBALBFD % (
                min_tx_interval_xml, min_rx_interval_xml, multiplier_xml)
            conf_str = build_config_xml(configxmlstr)
            recv_xml = set_nc_config(self.module, conf_str)
            self._checkresponse_(recv_xml, "OPERATE_STATIC_ROUTE_globalBFD")

        if self.state == "absent" and self.commands:
            min_tx_interval_xml = CE_NC_SET_IPV4_STATIC_ROUTE_BFDCOMMON_MINTX % 1000
            min_rx_interval_xml = CE_NC_SET_IPV4_STATIC_ROUTE_BFDCOMMON_MINRX % 1000
            multiplier_xml = CE_NC_SET_IPV4_STATIC_ROUTE_BFDCOMMON_MUL % 3

            configxmlstr = CE_NC_SET_IPV4_STATIC_ROUTE_GLOBALBFD % (
                min_tx_interval_xml, min_rx_interval_xml, multiplier_xml)
            conf_str = build_config_xml(configxmlstr)
            recv_xml = set_nc_config(self.module, conf_str)
            self._checkresponse_(recv_xml, "OPERATE_STATIC_ROUTE_globalBFD")

    def operate_static_route_singlebfd(self, version, prefix, nhp_interface, next_hop, destvrf, state):
        """operate ipv4 static route singleBFD"""
        min_tx_interval = self.min_tx_interval
        min_rx_interval = self.min_rx_interval
        multiplier = self.detect_multiplier
        min_tx_interval_xml = """\n"""
        min_rx_interval_xml = """\n"""
        multiplier_xml = """\n"""
        local_address_xml = """\n"""
        if next_hop is None:
            next_hop = '0.0.0.0'

        if destvrf is None:
            dest_vpn_instance = "_public_"
        else:
            dest_vpn_instance = destvrf

        if nhp_interface is None:
            nhp_interface = "Invalid0"

        if min_tx_interval is not None:
            min_tx_interval_xml = CE_NC_SET_IPV4_STATIC_ROUTE_BFDCOMMON_MINTX % min_tx_interval
        if min_rx_interval is not None:
            min_rx_interval_xml = CE_NC_SET_IPV4_STATIC_ROUTE_BFDCOMMON_MINRX % min_rx_interval
        if multiplier is not None:
            multiplier_xml = CE_NC_SET_IPV4_STATIC_ROUTE_BFDCOMMON_MUL % multiplier

        if prefix is not None:
            local_address_xml = CE_NC_SET_STATIC_ROUTE_SINGLEBFD_LOCALADRESS % prefix

        if state == "present":
            configxmlstr = CE_NC_SET_STATIC_ROUTE_SINGLEBFD % (
                version, nhp_interface, dest_vpn_instance,
                next_hop, local_address_xml, min_tx_interval_xml,
                min_rx_interval_xml, multiplier_xml)

        else:
            configxmlstr = CE_NC_DELETE_STATIC_ROUTE_SINGLEBFD % (
                version, nhp_interface, dest_vpn_instance, next_hop)

        conf_str = build_config_xml(configxmlstr)

        recv_xml = set_nc_config(self.module, conf_str)
        self._checkresponse_(recv_xml, "OPERATE_STATIC_ROUTE_singleBFD")

    def operate_static_route(self, version, prefix, mask, nhp_interface, next_hop, vrf, destvrf, state):
        """operate ipv4 static route"""
        description_xml = """\n"""
        preference_xml = """\n"""
        tag_xml = """\n"""
        bfd_xml = """\n"""
        if next_hop is None:
            next_hop = '0.0.0.0'
        if nhp_interface is None:
            nhp_interface = "Invalid0"

        if vrf is None:
            vpn_instance = "_public_"
        else:
            vpn_instance = vrf

        if destvrf is None:
            dest_vpn_instance = "_public_"
        else:
            dest_vpn_instance = destvrf

        description_xml = get_xml(CE_NC_SET_DESCRIPTION, self.description)

        preference_xml = get_xml(CE_NC_SET_PREFERENCE, self.pref)

        tag_xml = get_xml(CE_NC_SET_TAG, self.tag)

        if self.function_flag == 'staticBFD':
            if self.bfd_session_name:
                bfd_xml = CE_NC_SET_BFDSESSIONNAME % self.bfd_session_name
        else:
            bfd_xml = CE_NC_SET_BFDENABLE
        if state == "present":
            configxmlstr = CE_NC_SET_STATIC_ROUTE % (
                vpn_instance, version, prefix, mask, nhp_interface,
                dest_vpn_instance, next_hop, description_xml, preference_xml, tag_xml, bfd_xml)

        else:
            configxmlstr = CE_NC_DELETE_STATIC_ROUTE % (
                vpn_instance, version, prefix, mask, nhp_interface, dest_vpn_instance, next_hop)

        conf_str = build_config_xml(configxmlstr)
        recv_xml = set_nc_config(self.module, conf_str)
        self._checkresponse_(recv_xml, "OPERATE_STATIC_ROUTE")

    def get_change_state_global_bfd(self):
        """get ipv4 global bfd change state"""

        self.get_global_bfd(self.state)
        change = False
        if self.state == "present":
            if self.static_routes_info["sroute_global_bfd"]:
                for static_route in self.static_routes_info["sroute_global_bfd"]:
                    if static_route is not None:
                        if self.min_tx_interval is not None:
                            if int(static_route["minTxInterval"]) != self.min_tx_interval:
                                change = True
                        if self.min_rx_interval is not None:
                            if int(static_route["minRxInterval"]) != self.min_rx_interval:
                                change = True
                        if self.detect_multiplier is not None:
                            if int(static_route["multiplier"]) != self.detect_multiplier:
                                change = True
                        return change
                    else:
                        continue
            else:
                change = True
        else:
            if self.commands:
                if self.static_routes_info["sroute_global_bfd"]:
                    for static_route in self.static_routes_info["sroute_global_bfd"]:
                        if static_route is not None:
                            if int(static_route["minTxInterval"]) != 1000 or \
                               int(static_route["minRxInterval"]) != 1000 or \
                               int(static_route["multiplier"]) != 3:
                                change = True
            return change

    def get_global_bfd(self, state):
        """get ipv4 global bfd"""

        self.static_routes_info["sroute_global_bfd"] = list()

        getglobalbfdxmlstr = None
        if self.aftype == 'v4':
            getglobalbfdxmlstr = CE_NC_GET_STATIC_ROUTE_IPV4_GLOBAL_BFD

        if getglobalbfdxmlstr is not None:
            xml_global_bfd_str = get_nc_config(self.module, getglobalbfdxmlstr)

            if 'data/' in xml_global_bfd_str:
                return

            xml_global_bfd_str = xml_global_bfd_str.replace('\r', '').replace('\n', ''). \
                replace('xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"', ""). \
                replace('xmlns="http://www.huawei.com/netconf/vrp"', "")
            root = ElementTree.fromstring(xml_global_bfd_str)
            static_routes_global_bfd = root.findall(
                "staticrt/staticrtbase/srIPv4StaticSite")

            if static_routes_global_bfd:
                for static_route in static_routes_global_bfd:
                    static_info = dict()
                    for static_ele in static_route:
                        if static_ele.tag == "minTxInterval":
                            if static_ele.text is not None:
                                static_info["minTxInterval"] = static_ele.text
                        if static_ele.tag == "minRxInterval":
                            if static_ele.text is not None:
                                static_info["minRxInterval"] = static_ele.text
                        if static_ele.tag == "multiplier":
                            if static_ele.text is not None:
                                static_info["multiplier"] = static_ele.text

                    self.static_routes_info["sroute_global_bfd"].append(static_info)

    def get_change_state_single_bfd(self):
        """get ipv4 single bfd change state"""

        self.get_single_bfd(self.state)
        change = False
        version = self.version
        if self.state == 'present':
            if self.static_routes_info["sroute_single_bfd"]:
                for static_route in self.static_routes_info["sroute_single_bfd"]:
                    if static_route is not None and static_route['afType'] == version:
                        if self.nhp_interface:
                            if static_route["ifName"].lower() != self.nhp_interface.lower():
                                change = True
                        if self.destvrf:
                            if static_route["destVrfName"].lower() != self.destvrf.lower():
                                change = True
                        if self.next_hop:
                            if static_route["nexthop"].lower() != self.next_hop.lower():
                                change = True
                        if self.prefix:
                            if static_route["localAddress"].lower() != self.prefix.lower():
                                change = True
                        if self.min_tx_interval:
                            if int(static_route["minTxInterval"]) != self.min_tx_interval:
                                change = True
                        if self.min_rx_interval:
                            if int(static_route["minRxInterval"]) != self.min_rx_interval:
                                change = True
                        if self.detect_multiplier:
                            if int(static_route["multiplier"]) != self.detect_multiplier:
                                change = True
                        return change

                    else:
                        continue
            else:
                change = True
        else:
            for static_route in self.static_routes_info["sroute_single_bfd"]:
                # undo ip route-static bfd [ interface-type interface-number |
                # vpn-instance vpn-instance-name ] nexthop-address

                if static_route["ifName"] and self.nhp_interface:
                    if static_route["ifName"].lower() == self.nhp_interface.lower() \
                            and static_route["nexthop"].lower() == self.next_hop.lower() \
                            and static_route["afType"] == version:
                        change = True
                        return change

                if static_route["destVrfName"] and self.destvrf:
                    if static_route["destVrfName"].lower() == self.destvrf.lower() \
                            and static_route["nexthop"].lower() == self.next_hop.lower() \
                            and static_route["afType"] == version:
                        change = True
                        return change

                if static_route["nexthop"] and self.next_hop:
                    if static_route["nexthop"].lower() == self.next_hop.lower() \
                            and static_route["afType"] == version:
                        change = True
                        return change
                else:
                    continue
            change = False
        return change

    def get_single_bfd(self, state):
        """get ipv4 sigle bfd"""
        self.static_routes_info["sroute_single_bfd"] = list()
        if self.aftype == "v4":
            version = "ipv4unicast"
        else:
            version = "ipv6unicast"
        if state == 'absent':
            getbfdxmlstr = CE_NC_GET_STATIC_ROUTE_BFD_ABSENT % (
                version, self.nhp_interface, self.destvrf, self.next_hop)
        else:
            getbfdxmlstr = CE_NC_GET_STATIC_ROUTE_BFD % (
                version, self.nhp_interface, self.destvrf, self.next_hop)
        xml_bfd_str = get_nc_config(self.module, getbfdxmlstr)

        if 'data/' in xml_bfd_str:
            return
        xml_bfd_str = xml_bfd_str.replace('\r', '').replace('\n', ''). \
            replace('xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"', ""). \
            replace('xmlns="http://www.huawei.com/netconf/vrp"', "")
        root = ElementTree.fromstring(xml_bfd_str)
        static_routes_bfd = root.findall(
            "staticrt/staticrtbase/srBfdParas/srBfdPara")
        if static_routes_bfd:
            for static_route in static_routes_bfd:
                static_info = dict()
                for static_ele in static_route:
                    if static_ele.tag in ["afType", "destVrfName", "nexthop", "ifName"]:
                        static_info[static_ele.tag] = static_ele.text
                    if static_ele.tag == "localAddress":
                        if static_ele.text is not None:
                            static_info["localAddress"] = static_ele.text
                        else:
                            static_info["localAddress"] = "None"
                    if static_ele.tag == "minTxInterval":
                        if static_ele.text is not None:
                            static_info["minTxInterval"] = static_ele.text
                    if static_ele.tag == "minRxInterval":
                        if static_ele.text is not None:
                            static_info["minRxInterval"] = static_ele.text
                    if static_ele.tag == "multiplier":
                        if static_ele.text is not None:
                            static_info["multiplier"] = static_ele.text
                self.static_routes_info["sroute_single_bfd"].append(static_info)

    def get_static_route(self, state):
        """get ipv4 static route about BFD"""
        self.static_routes_info["sroute"] = list()
        # Increase the parameter used to distinguish whether the incoming bfdSessionName
        static_bfd_flag = True
        if self.bfd_session_name:
            static_bfd_flag = False

        if state == 'absent':
            getxmlstr = CE_NC_GET_STATIC_ROUTE_ABSENT
        else:
            # self.static_bfd_flag is true
            if static_bfd_flag:
                getxmlstr = CE_NC_GET_STATIC_ROUTE_BFD_ENABLE

            else:
                getxmlstr = CE_NC_GET_STATIC_ROUTE_BFD_SESSIONNAME
        xml_str = get_nc_config(self.module, getxmlstr)
        if 'data/' in xml_str:
            return
        xml_str = xml_str.replace('\r', '').replace('\n', ''). \
            replace('xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"', ""). \
            replace('xmlns="http://www.huawei.com/netconf/vrp"', "")
        root = ElementTree.fromstring(xml_str)
        static_routes = root.findall(
            "staticrt/staticrtbase/srRoutes/srRoute")

        if static_routes:
            for static_route in static_routes:
                static_info = dict()
                for static_ele in static_route:
                    if static_ele.tag in ["vrfName", "afType", "topologyName",
                                          "prefix", "maskLength", "destVrfName",
                                          "nexthop", "ifName", "preference", "description"]:
                        static_info[static_ele.tag] = static_ele.text
                    if static_ele.tag == "tag":
                        if static_ele.text is not None:
                            static_info["tag"] = static_ele.text
                        else:
                            static_info["tag"] = "None"
                    if static_bfd_flag:
                        if static_ele.tag == "bfdEnable":
                            if static_ele.text is not None:
                                static_info["bfdEnable"] = static_ele.text
                            else:
                                static_info["bfdEnable"] = "None"
                    else:
                        if static_ele.tag == "sessionName":
                            if static_ele.text is not None:
                                static_info["sessionName"] = static_ele.text
                            else:
                                static_info["sessionName"] = "None"
                self.static_routes_info["sroute"].append(static_info)

    def _checkparams_(self):
        """check all input params"""
        if self.function_flag == 'singleBFD':
            if not self.next_hop:
                self.module.fail_json(msg='Error: missing required argument: next_hop.')
            if self.state != 'absent':
                if self.nhp_interface == "Invalid0" and (not self.prefix or self.prefix == '0.0.0.0'):
                    self.module.fail_json(msg='Error: If a nhp_interface is not configured, '
                                              'the prefix must be configured.')

        if self.function_flag != 'globalBFD':
            if self.function_flag == 'dynamicBFD' or self.function_flag == 'staticBFD':
                if not self.mask:
                    self.module.fail_json(msg='Error: missing required argument: mask.')
                # check prefix and mask
                if not self.mask.isdigit():
                    self.module.fail_json(msg='Error: Mask is invalid.')
            if self.function_flag != 'singleBFD' or (self.function_flag == 'singleBFD' and self.destvrf != "_public_"):
                if not self.prefix:
                    self.module.fail_json(msg='Error: missing required argument: prefix.')
                # convert prefix
                if not self._convertipprefix_():
                    self.module.fail_json(msg='Error: The %s is not a valid address' % self.prefix)

            if self.nhp_interface != "Invalid0" and self.destvrf != "_public_":
                self.module.fail_json(msg='Error: Destination vrf dose not support next hop is interface.')

            if not self.next_hop and self.nhp_interface == "Invalid0":
                self.module.fail_json(msg='Error: one of the following is required: next_hop,nhp_interface.')

        if self.function_flag == 'dynamicBFD' or self.function_flag == 'staticBFD':
            # description check
            if self.description:
                if not is_valid_description(self.description):
                    self.module.fail_json(
                        msg='Error: Dsecription length should be 1 - 35, and can not contain "?".')
            # tag check
            if self.tag is not None:
                if not is_valid_tag(self.tag):
                    self.module.fail_json(
                        msg='Error: Tag should be integer 1 - 4294967295.')
            # preference check
            if self.pref is not None:
                if not is_valid_preference(self.pref):
                    self.module.fail_json(
                        msg='Error: Preference should be integer 1 - 255.')

            if self.function_flag == 'staticBFD':
                if self.bfd_session_name:
                    if not is_valid_bdf_session_name(self.bfd_session_name):
                        self.module.fail_json(
                            msg='Error: bfd_session_name length should be 1 - 15, and can not contain Space.')

        # ipv4 check
        if self.aftype == "v4":
            if self.function_flag == 'dynamicBFD' or self.function_flag == 'staticBFD':
                if int(self.mask) > 32 or int(self.mask) < 0:
                    self.module.fail_json(
                        msg='Error: Ipv4 mask must be an integer between 1 and 32.')
            # next_hop check
            if self.function_flag != 'globalBFD':
                if self.next_hop:
                    if not is_valid_v4addr(self.next_hop):
                        self.module.fail_json(
                            msg='Error: The %s is not a valid address.' % self.next_hop)
        # ipv6 check
        if self.aftype == "v6":
            if self.function_flag == 'dynamicBFD' or self.function_flag == 'staticBFD':
                if int(self.mask) > 128 or int(self.mask) < 0:
                    self.module.fail_json(
                        msg='Error: Ipv6 mask must be an integer between 1 and 128.')
            if self.function_flag != 'globalBFD':
                if self.next_hop:
                    if not is_valid_v6addr(self.next_hop):
                        self.module.fail_json(
                            msg='Error: The %s is not a valid address.' % self.next_hop)

        if self.function_flag == 'globalBFD' or self.function_flag == 'singleBFD':
            # BFD prarams
            if self.min_tx_interval:
                if not is_valid_bdf_interval(self.min_tx_interval):
                    self.module.fail_json(
                        msg='Error: min_tx_interval should be integer 50 - 1000.')
            if self.min_rx_interval:
                if not is_valid_bdf_interval(self.min_rx_interval):
                    self.module.fail_json(
                        msg='Error: min_rx_interval should be integer 50 - 1000.')
            if self.detect_multiplier:
                if not is_valid_bdf_multiplier(self.detect_multiplier):
                    self.module.fail_json(
                        msg='Error: detect_multiplier should be integer 3 - 50.')

            if self.function_flag == 'globalBFD':
                if self.state != 'absent':
                    if not self.min_tx_interval and not self.min_rx_interval and not self.detect_multiplier:
                        self.module.fail_json(
                            msg='Error: one of the following is required: min_tx_interval,'
                                'detect_multiplier,min_rx_interval.')
                else:
                    if not self.commands:
                        self.module.fail_json(
                            msg='Error: missing required argument: command.')
                    if compare_command(self.commands):
                        self.module.fail_json(
                            msg='Error: The command %s line is incorrect.' % (',').join(self.commands))

    def set_ip_static_route_globalbfd(self):
        """set ip static route globalBFD"""
        if not self.changed:
            return
        if self.aftype == "v4":
            self.operate_static_route_globalbfd()

    def set_ip_static_route_singlebfd(self):
        """set ip static route singleBFD"""
        if not self.changed:
            return
        version = None
        if self.aftype == "v4":
            version = "ipv4unicast"
        else:
            version = "ipv6unicast"
        self.operate_static_route_singlebfd(version, self.prefix, self.nhp_interface,
                                            self.next_hop, self.destvrf, self.state)

    def set_ip_static_route(self):
        """set ip static route"""
        if not self.changed:
            return
        version = None
        if self.aftype == "v4":
            version = "ipv4unicast"
        else:
            version = "ipv6unicast"
        self.operate_static_route(version, self.prefix, self.mask, self.nhp_interface,
                                  self.next_hop, self.vrf, self.destvrf, self.state)

    def is_prefix_exist(self, static_route, version):
        """is prefix mask nex_thop exist"""
        if static_route is None:
            return False
        if self.next_hop and self.nhp_interface:
            return static_route["prefix"].lower() == self.prefix.lower() \
                and static_route["maskLength"] == self.mask \
                and static_route["afType"] == version \
                and static_route["ifName"].lower() == self.nhp_interface.lower() \
                and static_route["nexthop"].lower() == self.next_hop.lower()

        if self.next_hop and not self.nhp_interface:
            return static_route["prefix"].lower() == self.prefix.lower() \
                and static_route["maskLength"] == self.mask \
                and static_route["afType"] == version \
                and static_route["nexthop"].lower() == self.next_hop.lower()

        if not self.next_hop and self.nhp_interface:
            return static_route["prefix"].lower() == self.prefix.lower() \
                and static_route["maskLength"] == self.mask \
                and static_route["afType"] == version \
                and static_route["ifName"].lower() == self.nhp_interface.lower()

    def get_ip_static_route(self):
        """get ip static route"""
        change = False
        version = self.version
        self.get_static_route(self.state)
        change_list = list()
        if self.state == 'present':
            for static_route in self.static_routes_info["sroute"]:
                if self.is_prefix_exist(static_route, self.version):
                    info_dict = dict()
                    exist_dict = dict()
                    if self.vrf:
                        info_dict["vrfName"] = self.vrf
                        exist_dict["vrfName"] = static_route["vrfName"]
                    if self.destvrf:
                        info_dict["destVrfName"] = self.destvrf
                        exist_dict["destVrfName"] = static_route["destVrfName"]
                    if self.description:
                        info_dict["description"] = self.description
                        exist_dict["description"] = static_route["description"]
                    if self.tag:
                        info_dict["tag"] = self.tag
                        exist_dict["tag"] = static_route["tag"]
                    if self.pref:
                        info_dict["preference"] = str(self.pref)
                        exist_dict["preference"] = static_route["preference"]
                    if self.nhp_interface:
                        if self.nhp_interface.lower() == "invalid0":
                            info_dict["ifName"] = "Invalid0"
                        else:
                            info_dict["ifName"] = "Invalid0"
                        exist_dict["ifName"] = static_route["ifName"]
                    if self.next_hop:
                        info_dict["nexthop"] = self.next_hop
                        exist_dict["nexthop"] = static_route["nexthop"]

                    if self.bfd_session_name:
                        info_dict["bfdEnable"] = 'true'

                    else:
                        info_dict["bfdEnable"] = 'false'
                    exist_dict["bfdEnable"] = static_route["bfdEnable"]

                    if exist_dict != info_dict:
                        change = True
                    else:
                        change = False
                    change_list.append(change)

            if False in change_list:
                change = False
            else:
                change = True
            return change

        else:
            for static_route in self.static_routes_info["sroute"]:
                if static_route["nexthop"] and self.next_hop:
                    if static_route["prefix"].lower() == self.prefix.lower() \
                            and static_route["maskLength"] == self.mask \
                            and static_route["nexthop"].lower() == self.next_hop.lower() \
                            and static_route["afType"] == version:
                        change = True
                        return change
                if static_route["ifName"] and self.nhp_interface:
                    if static_route["prefix"].lower() == self.prefix.lower() \
                            and static_route["maskLength"] == self.mask \
                            and static_route["ifName"].lower() == self.nhp_interface.lower() \
                            and static_route["afType"] == version:
                        change = True
                        return change
                else:
                    continue
            change = False
        return change

    def get_proposed(self):
        """get proposed information"""
        self.proposed['afType'] = self.aftype
        self.proposed['state'] = self.state
        if self.function_flag != 'globalBFD':
            self.proposed['ifName'] = self.nhp_interface
            self.proposed['destVrfName'] = self.destvrf
            self.proposed['next_hop'] = self.next_hop

        if self.function_flag == 'singleBFD':
            if self.prefix:
                self.proposed['localAddress'] = self.prefix

        if self.function_flag == 'globalBFD' or self.function_flag == 'singleBFD':
            self.proposed['minTxInterval'] = self.min_tx_interval
            self.proposed['minRxInterval'] = self.min_rx_interval
            self.proposed['multiplier'] = self.detect_multiplier

        if self.function_flag != 'globalBFD' and self.function_flag != 'singleBFD':
            self.proposed['prefix'] = self.prefix
            self.proposed['mask'] = self.mask
            self.proposed['vrfName'] = self.vrf
            if self.tag:
                self.proposed['tag'] = self.tag
            if self.description:
                self.proposed['description'] = self.description
            if self.pref is None:
                self.proposed['preference'] = 60
            else:
                self.proposed['preference'] = self.pref

            static_bfd_flag = True
            if self.bfd_session_name:
                static_bfd_flag = False
            if not static_bfd_flag:
                self.proposed['sessionName'] = self.bfd_session_name
            else:
                self.proposed['bfdEnable'] = 'true'

    def get_existing(self):
        """get existing information"""
        # globalBFD
        if self.function_flag == 'globalBFD':
            change = self.get_change_state_global_bfd()
            self.existing['sroute_global_bfd'] = self.static_routes_info["sroute_global_bfd"]
        # singleBFD
        elif self.function_flag == 'singleBFD':
            change = self.get_change_state_single_bfd()
            self.existing['sroute_single_bfd'] = self.static_routes_info["sroute_single_bfd"]
        # dynamicBFD / staticBFD
        else:
            change = self.get_ip_static_route()
            self.existing['static_sroute'] = self.static_routes_info["sroute"]
        self.changed = bool(change)

    def get_end_state(self):
        """get end state information"""

        # globalBFD
        if self.function_flag == 'globalBFD':
            self.get_global_bfd(self.state)
            self.end_state['sroute_global_bfd'] = self.static_routes_info["sroute_global_bfd"]
        # singleBFD
        elif self.function_flag == 'singleBFD':
            self.static_routes_info["sroute_single_bfd"] = list()
            self.get_single_bfd(self.state)
            self.end_state['sroute_single_bfd'] = self.static_routes_info["sroute_single_bfd"]
        # dynamicBFD / staticBFD
        else:
            self.get_static_route(self.state)
            self.end_state['static_sroute'] = self.static_routes_info["sroute"]

    def work(self):
        """worker"""
        self._checkparams_()
        self.get_existing()
        self.get_proposed()

        if self.function_flag == 'globalBFD':
            self.set_ip_static_route_globalbfd()
            self.set_update_cmd_globalbfd()
        elif self.function_flag == 'singleBFD':
            self.set_ip_static_route_singlebfd()
            self.set_update_cmd_singlebfd()
        else:
            self.set_ip_static_route()
            self.set_update_cmd()

        self.get_end_state()
        if self.existing == self.end_state:
            self.changed = False
        self.results['changed'] = self.changed
        self.results['proposed'] = self.proposed
        self.results['existing'] = self.existing
        self.results['end_state'] = self.end_state
        if self.changed:
            self.results['updates'] = self.updates_cmd
        else:
            self.results['updates'] = list()

        self.module.exit_json(**self.results)


def main():
    """main"""

    argument_spec = dict(
        prefix=dict(type='str'),
        mask=dict(type='str'),
        aftype=dict(choices=['v4', 'v6'], required=True),
        next_hop=dict(type='str'),
        nhp_interface=dict(type='str'),
        vrf=dict(type='str'),
        destvrf=dict(type='str'),
        tag=dict(type='int'),
        description=dict(type='str'),
        pref=dict(type='int'),
        # bfd
        function_flag=dict(required=True, choices=['globalBFD', 'singleBFD', 'dynamicBFD', 'staticBFD']),
        min_tx_interval=dict(type='int'),
        min_rx_interval=dict(type='int'),
        detect_multiplier=dict(type='int'),
        # bfd session name
        bfd_session_name=dict(type='str'),
        commands=dict(type='list', required=False),
        state=dict(choices=['absent', 'present'], default='present', required=False),
    )
    interface = StaticRouteBFD(argument_spec)
    interface.work()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = '''
---

module: ce_mdn_interface
version_added: '0.2.0'
short_description: Manages MDN configuration on HUAWEI CloudEngine switches.
description:
    - Manages MDN configuration on HUAWEI CloudEngine switches.
author: xuxiaowei0512 (@CloudEngine-Ansible)
options:
  lldpenable:
    description:
      - Set global LLDP enable state.
    type: str
    choices: ['enabled', 'disabled']
  mdnstatus:
    description:
      - Set interface MDN enable state.
    type: str
    choices: ['rxOnly', 'disabled']
  ifname:
    description:
      - Interface name.
    type: str
  state:
    description:
      - Manage the state of the resource.
    default: present
    type: str
    choices: ['present','absent']
notes:
  - This module requires the netconf system service be enabled on
    the remote device being managed.
  - This module works with connection C(netconf).
'''

EXAMPLES = '''
  - name: "Configure global LLDP enable state"
    community.network.ce_mdn_interface:
      lldpenable: enabled

  - name: "Configure interface MDN enable state"
    community.network.ce_mdn_interface:
      ifname: 10GE1/0/1
      mdnstatus: rxOnly
'''

RETURN = '''
proposed:
    description: k/v pairs of parameters passed into module
    returned: always
    type: dict
    sample: {
                "lldpenable": "enabled",
                "ifname": "10GE1/0/1",
                "mdnstatus": "rxOnly",
                "state":"present"
            }
existing:
    description: k/v pairs of existing global LLDP configration
    returned: always
    type: dict
    sample: {
                "lldpenable": "enabled",
                "ifname": "10GE1/0/1",
                "mdnstatus": "disabled"
            }
end_state:
    description: k/v pairs of global LLDP configration after module execution
    returned: always
    type: dict
    sample: {
                "lldpenable": "enabled",
                "ifname": "10GE1/0/1",
                "mdnstatus": "rxOnly"
            }
updates:
    description: command sent to the device
    returned: always
    type: list
    sample: [
                "interface 10ge 1/0/1",
                "lldp mdn enable",
            ]
changed:
    description: check to see if a change was made on the device
    returned: always
    type: bool
    sample: true
'''

import copy
import re
from xml.etree import ElementTree
from ansible.module_utils.basic import AnsibleModule
from ansible_collections.community.network.plugins.module_utils.network.cloudengine.ce import set_nc_config, get_nc_config, execute_nc_action

CE_NC_GET_GLOBAL_LLDPENABLE_CONFIG = """
<filter type="subtree">
  <lldp xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
    <lldpSys>
      <lldpEnable></lldpEnable>
    </lldpSys>
  </lldp>
</filter>
"""

CE_NC_MERGE_GLOBA_LLDPENABLE_CONFIG = """
<config>
  <lldp xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
    <lldpSys operation="merge">
      <lldpEnable>%s</lldpEnable>
    </lldpSys>
  </lldp>
</config>
"""

CE_NC_GET_INTERFACE_MDNENABLE_CONFIG = """
<filter type="subtree">
  <lldp xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
    <mdnInterfaces>
      <mdnInterface>
        <ifName></ifName>
        <mdnStatus></mdnStatus>
      </mdnInterface>
    </mdnInterfaces>
  </lldp>
</filter>
"""

CE_NC_MERGE_INTERFACE_MDNENABLE_CONFIG = """
<config>
  <lldp xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
    <mdnInterfaces>
      <mdnInterface operation="merge">
        <ifName>%s</ifName>
        <mdnStatus>%s</mdnStatus>
      </mdnInterface>
    </mdnInterfaces>
  </lldp>
</config>
"""


def get_interface_type(interface):
    """Gets the type of interface, such as 10GE, ..."""

    if interface is None:
        return None

    iftype = None

    if interface.upper().startswith('GE'):
        iftype = 'ge'
    elif interface.upper().startswith('10GE'):
        iftype = '10ge'
    elif interface.upper().startswith('25GE'):
        iftype = '25ge'
    elif interface.upper().startswith('40GE'):
        iftype = '40ge'
    elif interface.upper().startswith('100GE'):
        iftype = '100ge'
    elif interface.upper().startswith('PORT-GROUP'):
        iftype = 'stack-Port'
    elif interface.upper().startswith('NULL'):
        iftype = 'null'
    else:
        return None
    return iftype.lower()


class Interface_mdn(object):
    """Manage global lldp enable configration"""

    def __init__(self, argument_spec):
        self.spec = argument_spec
        self.module = None
        self.init_module()

        # LLDP global configration info
        self.lldpenable = self.module.params['lldpenable'] or None
        self.ifname = self.module.params['ifname']
        self.mdnstatus = self.module.params['mdnstatus'] or None
        self.state = self.module.params['state']
        self.lldp_conf = dict()
        self.conf_exsit = False
        self.enable_flag = 0
        self.check_params()

        # state
        self.changed = False
        self.proposed_changed = dict()
        self.updates_cmd = list()
        self.results = dict()
        self.proposed = dict()
        self.existing = dict()
        self.end_state = dict()

    def check_params(self):
        """Check all input params"""

        if self.ifname:
            intf_type = get_interface_type(self.ifname)
            if not intf_type:
                self.module.fail_json(
                    msg='Error: ifname name of %s '
                        'is error.' % self.ifname)
            if (len(self.ifname) < 1) or (len(self.ifname) > 63):
                self.module.fail_json(
                    msg='Error: Ifname length is beetween 1 and 63.')

    def init_module(self):
        """Init module object"""

        self.module = AnsibleModule(
            argument_spec=self.spec, supports_check_mode=True)

    def check_response(self, xml_str, xml_name):
        """Check if response message is already succeed"""

        if "<ok/>" not in xml_str:
            self.module.fail_json(msg='Error: %s failed.' % xml_name)

    def config_interface_mdn(self):
        """Configure lldp enabled and interface mdn enabled parameters"""

        if self.state == 'present':
            if self.enable_flag == 0 and self.lldpenable == 'enabled':
                xml_str = CE_NC_MERGE_GLOBA_LLDPENABLE_CONFIG % self.lldpenable
                ret_xml = set_nc_config(self.module, xml_str)
                self.check_response(ret_xml, "LLDP_ENABLE_CONFIG")
                self.changed = True
            elif self.enable_flag == 1 and self.lldpenable == 'disabled':
                xml_str = CE_NC_MERGE_GLOBA_LLDPENABLE_CONFIG % self.lldpenable
                ret_xml = set_nc_config(self.module, xml_str)
                self.check_response(ret_xml, "LLDP_ENABLE_CONFIG")
                self.changed = True
            elif self.enable_flag == 1 and self.conf_exsit:
                xml_str = CE_NC_MERGE_INTERFACE_MDNENABLE_CONFIG % (self.ifname, self.mdnstatus)
                ret_xml = set_nc_config(self.module, xml_str)
                self.check_response(ret_xml, "INTERFACE_MDN_ENABLE_CONFIG")
                self.changed = True

    def show_result(self):
        """Show result"""

        self.results['changed'] = self.changed
        self.results['proposed'] = self.proposed
        self.results['existing'] = self.existing
        self.results['end_state'] = self.end_state
        if self.changed:
            self.results['updates'] = self.updates_cmd
        else:
            self.results['updates'] = list()

        self.module.exit_json(**self.results)

    def get_interface_mdn_exist_config(self):
        """Get lldp existed configure"""

        lldp_config = list()
        lldp_dict = dict()
        conf_enable_str = CE_NC_GET_GLOBAL_LLDPENABLE_CONFIG
        conf_enable_obj = get_nc_config(self.module, conf_enable_str)
        xml_enable_str = conf_enable_obj.replace('\r', '').replace('\n', '').\
            replace('xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"', "").\
            replace('xmlns="http://www.huawei.com/netconf/vrp"', "")

        # get lldp enable config info
        root_enable = ElementTree.fromstring(xml_enable_str)
        ntpsite_enable = root_enable.findall("lldp/lldpSys")
        for nexthop_enable in ntpsite_enable:
            for ele_enable in nexthop_enable:
                if ele_enable.tag in ["lldpEnable"]:
                    lldp_dict[ele_enable.tag] = ele_enable.text

            if self.state == "present":
                if lldp_dict['lldpEnable'] == 'enabled':
                    self.enable_flag = 1
            lldp_config.append(dict(lldpenable=lldp_dict['lldpEnable']))

        if self.enable_flag == 1:
            conf_str = CE_NC_GET_INTERFACE_MDNENABLE_CONFIG
            conf_obj = get_nc_config(self.module, conf_str)
            if "<data/>" in conf_obj:
                return lldp_config
            xml_str = conf_obj.replace('\r', '').replace('\n', '').\
                replace('xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"', "").\
                replace('xmlns="http://www.huawei.com/netconf/vrp"', "")
            # get all ntp config info
            root = ElementTree.fromstring(xml_str)
            ntpsite = root.findall("lldp/mdnInterfaces/mdnInterface")
            for nexthop in ntpsite:
                for ele in nexthop:
                    if ele.tag in ["ifName", "mdnStatus"]:
                        lldp_dict[ele.tag] = ele.text
                if self.state == "present":
                    cur_interface_mdn_cfg = dict(ifname=lldp_dict['ifName'], mdnstatus=lldp_dict['mdnStatus'])
                    exp_interface_mdn_cfg = dict(ifname=self.ifname, mdnstatus=self.mdnstatus)
                    if self.ifname == lldp_dict['ifName']:
                        if cur_interface_mdn_cfg != exp_interface_mdn_cfg:
                            self.conf_exsit = True
                            lldp_config.append(dict(ifname=lldp_dict['ifName'], mdnstatus=lldp_dict['mdnStatus']))
                            return lldp_config
                        lldp_config.append(dict(ifname=lldp_dict['ifName'], mdnstatus=lldp_dict['mdnStatus']))
        return lldp_config

    def get_existing(self):
        """Get existing info"""

        self.existing = self.get_interface_mdn_exist_config()

    def get_proposed(self):
        """Get proposed info"""

        if self.lldpenable:
            self.proposed = dict(lldpenable=self.lldpenable)
        if self.enable_flag == 1:
            if self.ifname:
                self.proposed = dict(ifname=self.ifname, mdnstatus=self.mdnstatus)

    def get_end_state(self):
        """Get end state info"""

        self.end_state = self.get_interface_mdn_exist_config()

    def get_update_cmd(self):
        """Get updated commands"""

        update_list = list()
        if self.state == "present":
            if self.lldpenable == "enabled":
                cli_str = "lldp enable"
                update_list.append(cli_str)
                if self.ifname:
                    cli_str = "%s %s" % ("interface", self.ifname)
                    update_list.append(cli_str)
                if self.mdnstatus:
                    if self.mdnstatus == "rxOnly":
                        cli_str = "lldp mdn enable"
                        update_list.append(cli_str)
                    else:
                        cli_str = "undo lldp mdn enable"
                        update_list.append(cli_str)

            elif self.lldpenable == "disabled":
                cli_str = "undo lldp enable"
                update_list.append(cli_str)
            else:
                if self.enable_flag == 1:
                    if self.ifname:
                        cli_str = "%s %s" % ("interface", self.ifname)
                        update_list.append(cli_str)
                    if self.mdnstatus:
                        if self.mdnstatus == "rxOnly":
                            cli_str = "lldp mdn enable"
                            update_list.append(cli_str)
                        else:
                            cli_str = "undo lldp mdn enable"
                            update_list.append(cli_str)

        self.updates_cmd.append(update_list)

    def work(self):
        """Excute task"""
        self.check_params()
        self.get_existing()
        self.get_proposed()
        self.config_interface_mdn()
        self.get_update_cmd()
        self.get_end_state()
        self.show_result()


def main():
    """Main function entry"""

    argument_spec = dict(
        lldpenable=dict(type='str', choices=['enabled', 'disabled']),
        mdnstatus=dict(type='str', choices=['rxOnly', 'disabled']),
        ifname=dict(type='str'),
        state=dict(choices=['absent', 'present'], default='present'),
    )
    lldp_obj = Interface_mdn(argument_spec)
    lldp_obj.work()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2020, quidame <quidame@poivron.org>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = r'''
---
module: iptables_state
short_description: Save iptables state into a file or restore it from a file
version_added: '1.1.0'
author: quidame (@quidame)
description:
  - C(iptables) is used to set up, maintain, and inspect the tables of IP
    packet filter rules in the Linux kernel.
  - This module handles the saving and/or loading of rules. This is the same
    as the behaviour of the C(iptables-save) and C(iptables-restore) (or
    C(ip6tables-save) and C(ip6tables-restore) for IPv6) commands which this
    module uses internally.
  - Modifying the state of the firewall remotely may lead to loose access to
    the host in case of mistake in new ruleset. This module embeds a rollback
    feature to avoid this, by telling the host to restore previous rules if a
    cookie is still there after a given delay, and all this time telling the
    controller to try to remove this cookie on the host through a new
    connection.
notes:
  - The rollback feature is not a module option and depends on task's
    attributes. To enable it, the module must be played asynchronously, i.e.
    by setting task attributes I(poll) to C(0), and I(async) to a value less
    or equal to C(ANSIBLE_TIMEOUT). If I(async) is greater, the rollback will
    still happen if it shall happen, but you will experience a connection
    timeout instead of more relevant info returned by the module after its
    failure.
  - This module supports I(check_mode).
options:
  counters:
    description:
      - Save or restore the values of all packet and byte counters.
      - When C(true), the module is not idempotent.
    type: bool
    default: false
  ip_version:
    description:
      - Which version of the IP protocol this module should apply to.
    type: str
    choices: [ ipv4, ipv6 ]
    default: ipv4
  modprobe:
    description:
      - Specify the path to the C(modprobe) program internally used by iptables
        related commands to load kernel modules.
      - By default, C(/proc/sys/kernel/modprobe) is inspected to determine the
        executable's path.
    type: path
  noflush:
    description:
      - For I(state=restored), ignored otherwise.
      - If C(false), restoring iptables rules from a file flushes (deletes)
        all previous contents of the respective table(s). If C(true), the
        previous rules are left untouched (but policies are updated anyway,
        for all built-in chains).
    type: bool
    default: false
  path:
    description:
      - The file the iptables state should be saved to.
      - The file the iptables state should be restored from.
    type: path
    required: yes
  state:
    description:
      - Whether the firewall state should be saved (into a file) or restored
        (from a file).
    type: str
    choices: [ saved, restored ]
    required: yes
  table:
    description:
      - When I(state=restored), restore only the named table even if the input
        file contains other tables. Fail if the named table is not declared in
        the file.
      - When I(state=saved), restrict output to the specified table. If not
        specified, output includes all active tables.
    type: str
    choices: [ filter, nat, mangle, raw, security ]
  wait:
    description:
      - Wait N seconds for the xtables lock to prevent instant failure in case
        multiple instances of the program are running concurrently.
    type: int
requirements: [iptables, ip6tables]
'''

EXAMPLES = r'''
# This will apply to all loaded/active IPv4 tables.
- name: Save current state of the firewall in system file
  community.general.iptables_state:
    state: saved
    path: /etc/sysconfig/iptables

# This will apply only to IPv6 filter table.
- name: save current state of the firewall in system file
  community.general.iptables_state:
    ip_version: ipv6
    table: filter
    state: saved
    path: /etc/iptables/rules.v6

# This will load a state from a file, with a rollback in case of access loss
- name: restore firewall state from a file
  community.general.iptables_state:
    state: restored
    path: /run/iptables.apply
  async: "{{ ansible_timeout }}"
  poll: 0

# This will load new rules by appending them to the current ones
- name: restore firewall state from a file
  community.general.iptables_state:
    state: restored
    path: /run/iptables.apply
    noflush: true
  async: "{{ ansible_timeout }}"
  poll: 0

# This will only retrieve information
- name: get current state of the firewall
  community.general.iptables_state:
    state: saved
    path: /tmp/iptables
  check_mode: yes
  changed_when: false
  register: iptables_state

- name: show current state of the firewall
  ansible.builtin.debug:
    var: iptables_state.initial_state
'''

RETURN = r'''
applied:
  description: Whether or not the wanted state has been successfully restored.
  type: bool
  returned: always
  sample: true
initial_state:
  description: The current state of the firewall when module starts.
  type: list
  elements: str
  returned: always
  sample: [
      "# Generated by xtables-save v1.8.2",
      "*filter",
      ":INPUT ACCEPT [0:0]",
      ":FORWARD ACCEPT [0:0]",
      ":OUTPUT ACCEPT [0:0]",
      "COMMIT",
      "# Completed"
    ]
restored:
  description: The state the module restored, whenever it is finally applied or not.
  type: list
  elements: str
  returned: always
  sample: [
      "# Generated by xtables-save v1.8.2",
      "*filter",
      ":INPUT DROP [0:0]",
      ":FORWARD DROP [0:0]",
      ":OUTPUT ACCEPT [0:0]",
      "-A INPUT -m conntrack --ctstate RELATED,ESTABLISHED -j ACCEPT",
      "-A INPUT -m conntrack --ctstate INVALID -j DROP",
      "-A INPUT -i lo -j ACCEPT",
      "-A INPUT -p icmp -j ACCEPT",
      "-A INPUT -p tcp -m tcp --dport 22 -j ACCEPT",
      "COMMIT",
      "# Completed"
    ]
saved:
  description: The iptables state the module saved.
  type: list
  elements: str
  returned: always
  sample: [
      "# Generated by xtables-save v1.8.2",
      "*filter",
      ":INPUT ACCEPT [0:0]",
      ":FORWARD DROP [0:0]",
      ":OUTPUT ACCEPT [0:0]",
      "COMMIT",
      "# Completed"
    ]
tables:
  description: The iptables we have interest for when module starts.
  type: dict
  contains:
    table:
      description: Policies and rules for all chains of the named table.
      type: list
      elements: str
  sample: |-
    {
      "filter": [
        ":INPUT ACCEPT",
        ":FORWARD ACCEPT",
        ":OUTPUT ACCEPT",
        "-A INPUT -i lo -j ACCEPT",
        "-A INPUT -p icmp -j ACCEPT",
        "-A INPUT -p tcp -m tcp --dport 22 -j ACCEPT",
        "-A INPUT -j REJECT --reject-with icmp-host-prohibited"
      ],
      "nat": [
        ":PREROUTING ACCEPT",
        ":INPUT ACCEPT",
        ":OUTPUT ACCEPT",
        ":POSTROUTING ACCEPT"
      ]
    }
  returned: always
'''


import re
import os
import time
import tempfile
import filecmp
import shutil

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils._text import to_bytes, to_native, to_text


IPTABLES = dict(
    ipv4='iptables',
    ipv6='ip6tables',
)

SAVE = dict(
    ipv4='iptables-save',
    ipv6='ip6tables-save',
)

RESTORE = dict(
    ipv4='iptables-restore',
    ipv6='ip6tables-restore',
)

TABLES = ['filter', 'mangle', 'nat', 'raw', 'security']


def read_state(b_path):
    '''
    Read a file and store its content in a variable as a list.
    '''
    with open(b_path, 'r') as f:
        text = f.read()
    lines = text.splitlines()
    while '' in lines:
        lines.remove('')
    return (lines)


def write_state(b_path, lines, changed):
    '''
    Write given contents to the given path, and return changed status.
    '''
    # Populate a temporary file
    tmpfd, tmpfile = tempfile.mkstemp()
    with os.fdopen(tmpfd, 'w') as f:
        for line in lines:
            f.write('%s\n' % line)

    # Prepare to copy temporary file to the final destination
    if not os.path.exists(b_path):
        b_destdir = os.path.dirname(b_path)
        destdir = to_native(b_destdir, errors='surrogate_or_strict')
        if b_destdir and not os.path.exists(b_destdir) and not module.check_mode:
            try:
                os.makedirs(b_destdir)
            except Exception as e:
                module.fail_json(
                    msg='Error creating %s. Error code: %s. Error description: %s' % (destdir, e[0], e[1]),
                    initial_state=lines)
        changed = True

    elif not filecmp.cmp(tmpfile, b_path):
        changed = True

    # Do it
    if changed and not module.check_mode:
        try:
            shutil.copyfile(tmpfile, b_path)
        except Exception as e:
            path = to_native(b_path, errors='surrogate_or_strict')
            module.fail_json(
                msg='Error saving state into %s. Error code: %s. Error description: %s' % (path, e[0], e[1]),
                initial_state=lines)

    return changed


def initialize_from_null_state(initializer, initcommand, table):
    '''
    This ensures iptables-state output is suitable for iptables-restore to roll
    back to it, i.e. iptables-save output is not empty. This also works for the
    iptables-nft-save alternative.
    '''
    if table is None:
        table = 'filter'

    tmpfd, tmpfile = tempfile.mkstemp()
    with os.fdopen(tmpfd, 'w') as f:
        f.write('*%s\nCOMMIT\n' % table)

    initializer.append(tmpfile)
    (rc, out, err) = module.run_command(initializer, check_rc=True)
    (rc, out, err) = module.run_command(initcommand, check_rc=True)
    return (rc, out, err)


def filter_and_format_state(string):
    '''
    Remove timestamps to ensure idempotence between runs. Also remove counters
    by default. And return the result as a list.
    '''
    string = re.sub('((^|\n)# (Generated|Completed)[^\n]*) on [^\n]*', '\\1', string)
    if not module.params['counters']:
        string = re.sub('[[][0-9]+:[0-9]+[]]', '[0:0]', string)
    lines = string.splitlines()
    while '' in lines:
        lines.remove('')
    return (lines)


def per_table_state(command, state):
    '''
    Convert raw iptables-save output into usable datastructure, for reliable
    comparisons between initial and final states.
    '''
    tables = dict()
    for t in TABLES:
        COMMAND = list(command)
        if '*%s' % t in state.splitlines():
            COMMAND.extend(['--table', t])
            (rc, out, err) = module.run_command(COMMAND, check_rc=True)
            out = re.sub('(^|\n)(# Generated|# Completed|[*]%s|COMMIT)[^\n]*' % t, '', out)
            out = re.sub(' *[[][0-9]+:[0-9]+[]] *', '', out)
            table = out.splitlines()
            while '' in table:
                table.remove('')
            tables[t] = table
    return (tables)


def main():

    global module

    module = AnsibleModule(
        argument_spec=dict(
            path=dict(type='path', required=True),
            state=dict(type='str', choices=['saved', 'restored'], required=True),
            table=dict(type='str', choices=['filter', 'nat', 'mangle', 'raw', 'security']),
            noflush=dict(type='bool', default=False),
            counters=dict(type='bool', default=False),
            modprobe=dict(type='path'),
            ip_version=dict(type='str', choices=['ipv4', 'ipv6'], default='ipv4'),
            wait=dict(type='int'),
            _timeout=dict(type='int'),
            _back=dict(type='path'),
        ),
        required_together=[
            ['_timeout', '_back'],
        ],
        supports_check_mode=True,
    )

    # We'll parse iptables-restore stderr
    module.run_command_environ_update = dict(LANG='C', LC_MESSAGES='C')

    path = module.params['path']
    state = module.params['state']
    table = module.params['table']
    noflush = module.params['noflush']
    counters = module.params['counters']
    modprobe = module.params['modprobe']
    ip_version = module.params['ip_version']
    wait = module.params['wait']
    _timeout = module.params['_timeout']
    _back = module.params['_back']

    bin_iptables = module.get_bin_path(IPTABLES[ip_version], True)
    bin_iptables_save = module.get_bin_path(SAVE[ip_version], True)
    bin_iptables_restore = module.get_bin_path(RESTORE[ip_version], True)

    os.umask(0o077)
    changed = False
    COMMANDARGS = []
    INITCOMMAND = [bin_iptables_save]
    INITIALIZER = [bin_iptables_restore]
    TESTCOMMAND = [bin_iptables_restore, '--test']

    if counters:
        COMMANDARGS.append('--counters')

    if table is not None:
        COMMANDARGS.extend(['--table', table])

    if wait is not None:
        TESTCOMMAND.extend(['--wait', '%s' % wait])

    if modprobe is not None:
        b_modprobe = to_bytes(modprobe, errors='surrogate_or_strict')
        if not os.path.exists(b_modprobe):
            module.fail_json(msg="modprobe %s not found" % modprobe)
        if not os.path.isfile(b_modprobe):
            module.fail_json(msg="modprobe %s not a file" % modprobe)
        if not os.access(b_modprobe, os.R_OK):
            module.fail_json(msg="modprobe %s not readable" % modprobe)
        if not os.access(b_modprobe, os.X_OK):
            module.fail_json(msg="modprobe %s not executable" % modprobe)
        COMMANDARGS.extend(['--modprobe', modprobe])
        INITIALIZER.extend(['--modprobe', modprobe])
        INITCOMMAND.extend(['--modprobe', modprobe])
        TESTCOMMAND.extend(['--modprobe', modprobe])

    SAVECOMMAND = list(COMMANDARGS)
    SAVECOMMAND.insert(0, bin_iptables_save)

    b_path = to_bytes(path, errors='surrogate_or_strict')

    if state == 'restored':
        if not os.path.exists(b_path):
            module.fail_json(msg="Source %s not found" % path)
        if not os.path.isfile(b_path):
            module.fail_json(msg="Source %s not a file" % path)
        if not os.access(b_path, os.R_OK):
            module.fail_json(msg="Source %s not readable" % path)
        state_to_restore = read_state(b_path)
    else:
        cmd = ' '.join(SAVECOMMAND)

    (rc, stdout, stderr) = module.run_command(INITCOMMAND, check_rc=True)

    # The issue comes when wanting to restore state from empty iptable-save's
    # output... what happens when, say:
    # - no table is specified, and iptables-save's output is only nat table;
    # - we give filter's ruleset to iptables-restore, that locks ourselve out
    #   of the host;
    # then trying to roll iptables state back to the previous (working) setup
    # doesn't override current filter table because no filter table is stored
    # in the backup ! So we have to ensure tables to be restored have a backup
    # in case of rollback.
    if table is None:
        if state == 'restored':
            for t in TABLES:
                if '*%s' % t in state_to_restore:
                    if len(stdout) == 0 or '*%s' % t not in stdout.splitlines():
                        (rc, stdout, stderr) = initialize_from_null_state(INITIALIZER, INITCOMMAND, t)
        elif len(stdout) == 0:
            (rc, stdout, stderr) = initialize_from_null_state(INITIALIZER, INITCOMMAND, 'filter')

    elif state == 'restored' and '*%s' % table not in state_to_restore:
        module.fail_json(msg="Table %s to restore not defined in %s" % (table, path))

    elif len(stdout) == 0 or '*%s' % table not in stdout.splitlines():
        (rc, stdout, stderr) = initialize_from_null_state(INITIALIZER, INITCOMMAND, table)

    initial_state = filter_and_format_state(stdout)
    if initial_state is None:
        module.fail_json(msg="Unable to initialize firewall from NULL state.")

    # Depending on the value of 'table', initref_state may differ from
    # initial_state.
    (rc, stdout, stderr) = module.run_command(SAVECOMMAND, check_rc=True)
    tables_before = per_table_state(SAVECOMMAND, stdout)
    initref_state = filter_and_format_state(stdout)

    if state == 'saved':
        changed = write_state(b_path, initref_state, changed)
        module.exit_json(
            changed=changed,
            cmd=cmd,
            tables=tables_before,
            initial_state=initial_state,
            saved=initref_state)

    #
    # All remaining code is for state=restored
    #

    MAINCOMMAND = list(COMMANDARGS)
    MAINCOMMAND.insert(0, bin_iptables_restore)

    if wait is not None:
        MAINCOMMAND.extend(['--wait', '%s' % wait])

    if _back is not None:
        b_back = to_bytes(_back, errors='surrogate_or_strict')
        garbage = write_state(b_back, initref_state, changed)
        BACKCOMMAND = list(MAINCOMMAND)
        BACKCOMMAND.append(_back)

    if noflush:
        MAINCOMMAND.append('--noflush')

    MAINCOMMAND.append(path)
    cmd = ' '.join(MAINCOMMAND)

    TESTCOMMAND = list(MAINCOMMAND)
    TESTCOMMAND.insert(1, '--test')
    error_msg = "Source %s is not suitable for input to %s" % (path, os.path.basename(bin_iptables_restore))

    # Due to a bug in iptables-nft-restore --test, we have to validate tables
    # one by one (https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=960003).
    for t in tables_before:
        testcommand = list(TESTCOMMAND)
        testcommand.extend(['--table', t])
        (rc, stdout, stderr) = module.run_command(testcommand)

        if 'Another app is currently holding the xtables lock' in stderr:
            error_msg = stderr

        if rc != 0:
            cmd = ' '.join(testcommand)
            module.fail_json(
                msg=error_msg,
                cmd=cmd,
                rc=rc,
                stdout=stdout,
                stderr=stderr,
                tables=tables_before,
                initial_state=initial_state,
                restored=state_to_restore,
                applied=False)

    if module.check_mode:
        tmpfd, tmpfile = tempfile.mkstemp()
        with os.fdopen(tmpfd, 'w') as f:
            for line in initial_state:
                f.write('%s\n' % line)

        if filecmp.cmp(tmpfile, b_path):
            restored_state = initial_state
        else:
            restored_state = state_to_restore

    else:
        # Let time enough to the plugin to retrieve async status of the module
        # in case of bad option type/value and the like.
        if _back is not None:
            b_starter = to_bytes('%s.starter' % _back, errors='surrogate_or_strict')
            while True:
                if os.path.exists(b_starter):
                    os.remove(b_starter)
                    break
                else:
                    time.sleep(0.01)
                    continue

        (rc, stdout, stderr) = module.run_command(MAINCOMMAND)
        if 'Another app is currently holding the xtables lock' in stderr:
            module.fail_json(
                msg=stderr,
                cmd=cmd,
                rc=rc,
                stdout=stdout,
                stderr=stderr,
                tables=tables_before,
                initial_state=initial_state,
                restored=state_to_restore,
                applied=False)

        (rc, stdout, stderr) = module.run_command(SAVECOMMAND, check_rc=True)
        restored_state = filter_and_format_state(stdout)

    if restored_state != initref_state and restored_state != initial_state:
        if module.check_mode:
            changed = True
        else:
            tables_after = per_table_state(SAVECOMMAND, stdout)
            if tables_after != tables_before:
                changed = True

    if _back is None or module.check_mode:
        module.exit_json(
            changed=changed,
            cmd=cmd,
            tables=tables_before,
            initial_state=initial_state,
            restored=restored_state,
            applied=True)

    # The rollback implementation currently needs:
    # Here:
    # * test existence of the backup file, exit with success if it doesn't exist
    # * otherwise, restore iptables from this file and return failure
    # Action plugin:
    # * try to remove the backup file
    # * wait async task is finished and retrieve its final status
    # * modify it and return the result
    # Task:
    # * task attribute 'async' set to the same value (or lower) than ansible
    #   timeout
    # * task attribute 'poll' equals 0
    #
    for x in range(_timeout):
        if os.path.exists(b_back):
            time.sleep(1)
            continue
        module.exit_json(
            changed=changed,
            cmd=cmd,
            tables=tables_before,
            initial_state=initial_state,
            restored=restored_state,
            applied=True)

    # Here we are: for whatever reason, but probably due to the current ruleset,
    # the action plugin (i.e. on the controller) was unable to remove the backup
    # cookie, so we restore initial state from it.
    (rc, stdout, stderr) = module.run_command(BACKCOMMAND, check_rc=True)
    os.remove(b_back)

    (rc, stdout, stderr) = module.run_command(SAVECOMMAND, check_rc=True)
    tables_rollback = per_table_state(SAVECOMMAND, stdout)

    msg = (
        "Failed to confirm state restored from %s after %ss. "
        "Firewall has been rolled back to its initial state." % (path, _timeout)
    )

    module.fail_json(
        changed=(tables_before != tables_rollback),
        msg=msg,
        cmd=cmd,
        tables=tables_before,
        initial_state=initial_state,
        restored=restored_state,
        applied=False)


if __name__ == '__main__':
    main()



#
# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

"""
The nxos acl_interfaces fact class
It is in this file the configuration is collected from the device
for a given resource, parsed, and the facts tree is populated
based on the configuration.
"""
from __future__ import absolute_import, division, print_function

__metaclass__ = type

import re
from copy import deepcopy
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common import (
    utils,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.argspec.acl_interfaces.acl_interfaces import (
    Acl_interfacesArgs,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.utils.utils import (
    normalize_interface,
)


class Acl_interfacesFacts(object):
    """ The nxos acl_interfaces fact class
    """

    def __init__(self, module, subspec="config", options="options"):
        self._module = module
        self.argument_spec = Acl_interfacesArgs.argument_spec
        spec = deepcopy(self.argument_spec)
        if subspec:
            if options:
                facts_argument_spec = spec[subspec][options]
            else:
                facts_argument_spec = spec[subspec]
        else:
            facts_argument_spec = spec

        self.generated_spec = utils.generate_dict(facts_argument_spec)

    def get_device_data(self, connection):
        return connection.get("show running-config | section ^interface")

    def populate_facts(self, connection, ansible_facts, data=None):
        """ Populate the facts for acl_interfaces
        :param connection: the device connection
        :param ansible_facts: Facts dictionary
        :param data: previously collected conf
        :rtype: dictionary
        :returns: facts
        """
        if not data:
            data = self.get_device_data(connection)
        data = data.split("interface")

        resources = []
        for i in range(len(data)):
            intf = data[i].split("\n")
            for l in range(1, len(intf)):
                if not re.search(
                    "ip(v6)?( port)? (access-group|traffic-filter)", intf[l]
                ):
                    intf[l] = ""
            intf = list(filter(None, intf))
            resources.append(intf)

        objs = []
        for resource in resources:
            if resource:
                obj = self.render_config(self.generated_spec, resource)
                if obj:
                    objs.append(obj)

        ansible_facts["ansible_network_resources"].pop("acl_interfaces", None)
        facts = {}
        if objs:
            params = utils.validate_config(
                self.argument_spec, {"config": objs}
            )
            params = utils.remove_empties(params)
            facts["acl_interfaces"] = params["config"]

        ansible_facts["ansible_network_resources"].update(facts)
        return ansible_facts

    def render_config(self, spec, conf):
        """
        Render config as dictionary structure and delete keys
          from spec for null values

        :param spec: The facts tree, generated from the argspec
        :param conf: The configuration
        :rtype: dictionary
        :returns: The generated config
        """
        config = deepcopy(spec)
        config["name"] = conf[0].strip()
        config["access_groups"] = []
        v4 = {"afi": "ipv4", "acls": []}
        v6 = {"afi": "ipv6", "acls": []}
        for c in conf[1:]:
            if c:
                acl4 = re.search(
                    r"ip(?P<port>\sport)?\saccess-group\s(?P<name>\S+)\s(?P<dir>in|out)",
                    c,
                )
                acl6 = re.search(
                    r"ipv6(?P<port>\sport)?\straffic-filter\s(?P<name>\S+)\s(?P<dir>in|out)",
                    c,
                )
                if acl4:
                    v4["acls"].append(self._parse(acl4))
                elif acl6:
                    v6["acls"].append(self._parse(acl6))

        if len(v4["acls"]) > 0:
            config["access_groups"].append(v4)
        if len(v6["acls"]) > 0:
            config["access_groups"].append(v6)

        return utils.remove_empties(config)

    def _parse(self, data):
        return {
            "name": data.group("name").strip(),
            "direction": data.group("dir").strip(),
            "port": True if data.group("port") else None,
        }



#
# (c) 2019, Ansible by Red Hat, inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import absolute_import, division, print_function

__metaclass__ = type

from ansible_collections.cisco.nxos.plugins.modules import nxos_acls
from ansible_collections.cisco.nxos.tests.unit.compat.mock import (
    patch,
    MagicMock,
)
from ansible_collections.cisco.nxos.tests.unit.modules.utils import (
    set_module_args,
)
from .nxos_module import TestNxosModule, load_fixture


class TestNxosAclsModule(TestNxosModule):

    module = nxos_acls

    def setUp(self):
        super(TestNxosAclsModule, self).setUp()

        self.mock_get_config = patch(
            "ansible_collections.ansible.netcommon.plugins.module_utils.network.common.network.Config.get_config"
        )
        self.get_config = self.mock_get_config.start()

        self.mock_load_config = patch(
            "ansible_collections.ansible.netcommon.plugins.module_utils.network.common.network.Config.load_config"
        )
        self.load_config = self.mock_load_config.start()

        self.mock_get_resource_connection_config = patch(
            "ansible_collections.ansible.netcommon.plugins.module_utils.network.common.cfg.base.get_resource_connection"
        )
        self.get_resource_connection_config = (
            self.mock_get_resource_connection_config.start()
        )

        self.mock_get_resource_connection_facts = patch(
            "ansible_collections.ansible.netcommon.plugins.module_utils.network.common.facts.facts.get_resource_connection"
        )
        self.get_resource_connection_facts = (
            self.mock_get_resource_connection_facts.start()
        )

        self.mock_edit_config = patch(
            "ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.config.acls.acls.Acls.edit_config"
        )
        self.edit_config = self.mock_edit_config.start()

        self.mock_execute_show_command = patch(
            "ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.facts.acls.acls.AclsFacts.get_device_data"
        )
        self.execute_show_command = self.mock_execute_show_command.start()

    def tearDown(self):
        super(TestNxosAclsModule, self).tearDown()
        self.mock_get_resource_connection_config.stop()
        self.mock_get_resource_connection_facts.stop()
        self.mock_edit_config.stop()
        self.mock_get_config.stop()
        self.mock_load_config.stop()
        self.mock_execute_show_command.stop()

    def load_fixtures(self, commands=None, device=""):
        def load_from_file(*args, **kwargs):
            v4 = """\nip access-list ACL1v4\n 10 permit ip any any\n 20 deny udp any any"""
            v6 = """\nipv6 access-list ACL1v6\n 10 permit sctp any any"""
            return v4 + v6

        self.execute_show_command.side_effect = load_from_file

    def test_nxos_acls_merged(self):
        set_module_args(
            dict(
                config=[
                    dict(
                        afi="ipv4",
                        acls=[
                            dict(
                                name="ACL2v4",
                                aces=[
                                    dict(
                                        grant="deny",
                                        destination=dict(any=True),
                                        source=dict(any=True),
                                        fragments=True,
                                        sequence=20,
                                        protocol="tcp",
                                        protocol_options=dict(
                                            tcp=dict(ack=True)
                                        ),
                                    ),
                                    dict(
                                        grant="deny",
                                        destination=dict(
                                            prefix="2002:2:2:2::/64"
                                        ),
                                        source=dict(prefix="2002:1:1:1::/64"),
                                        sequence=30,
                                        protocol="icmp",
                                        protocol_options=dict(
                                            icmp=dict(echo_request=True)
                                        ),
                                    ),
                                ],
                            )
                        ],
                    ),
                    dict(afi="ipv6", acls=[dict(name="ACL2v6")]),
                ],
                state="merged",
            )
        )
        commands = [
            "ip access-list ACL2v4",
            "20 deny tcp any any ack fragments",
            "30 deny icmp 2002:1:1:1::/64 2002:2:2:2::/64 echo-request",
            "ipv6 access-list ACL2v6",
        ]
        self.execute_module(changed=True, commands=commands)

    def test_nxos_acls_merged_idempotent(self):
        set_module_args(
            dict(
                config=[
                    dict(
                        afi="ipv4",
                        acls=[
                            dict(
                                name="ACL1v4",
                                aces=[
                                    dict(
                                        grant="permit",
                                        destination=dict(any=True),
                                        source=dict(any=True),
                                        sequence=10,
                                        protocol="ip",
                                    ),
                                    dict(
                                        grant="deny",
                                        destination=dict(any=True),
                                        source=dict(any=True),
                                        sequence=20,
                                        protocol="udp",
                                    ),
                                ],
                            )
                        ],
                    ),
                    dict(
                        afi="ipv6",
                        acls=[
                            dict(
                                name="ACL1v6",
                                aces=[
                                    dict(
                                        grant="permit",
                                        destination=dict(any=True),
                                        source=dict(any=True),
                                        sequence=10,
                                        protocol="sctp",
                                    )
                                ],
                            )
                        ],
                    ),
                ],
                state="merged",
            )
        )
        self.execute_module(changed=False, commands=[])

    def test_nxos_acls_replaced(self):
        set_module_args(
            dict(
                config=[
                    dict(
                        afi="ipv4",
                        acls=[
                            dict(
                                name="ACL1v4",
                                aces=[
                                    dict(
                                        grant="permit",
                                        destination=dict(host="192.0.2.28"),
                                        source=dict(any=True),
                                        log=True,
                                        sequence=50,
                                        protocol="icmp",
                                        protocol_options=dict(
                                            icmp=dict(
                                                administratively_prohibited=True
                                            )
                                        ),
                                    )
                                ],
                            )
                        ],
                    )
                ],
                state="replaced",
            )
        )
        commands = [
            "ip access-list ACL1v4",
            "no 20 deny udp any any",
            "no 10 permit ip any any",
            "50 permit icmp any host 192.0.2.28 administratively-prohibited log",
        ]
        self.execute_module(changed=True, commands=commands)

    def test_nxos_acls_replaced_idempotent(self):
        set_module_args(
            dict(
                config=[
                    dict(
                        afi="ipv4",
                        acls=[
                            dict(
                                name="ACL1v4",
                                aces=[
                                    dict(
                                        grant="permit",
                                        destination=dict(any=True),
                                        source=dict(any=True),
                                        sequence=10,
                                        protocol="ip",
                                    ),
                                    dict(
                                        grant="deny",
                                        destination=dict(any=True),
                                        source=dict(any=True),
                                        sequence=20,
                                        protocol="udp",
                                    ),
                                ],
                            )
                        ],
                    ),
                    dict(
                        afi="ipv6",
                        acls=[
                            dict(
                                name="ACL1v6",
                                aces=[
                                    dict(
                                        grant="permit",
                                        destination=dict(any=True),
                                        source=dict(any=True),
                                        sequence=10,
                                        protocol="sctp",
                                    )
                                ],
                            )
                        ],
                    ),
                ],
                state="replaced",
            )
        )
        self.execute_module(changed=False, commands=[])

    def test_nxos_acls_overridden(self):
        set_module_args(
            dict(
                config=[
                    dict(
                        afi="ipv4",
                        acls=[
                            dict(
                                name="ACL2v4",
                                aces=[
                                    dict(
                                        grant="permit",
                                        destination=dict(host="192.0.2.28"),
                                        source=dict(any=True),
                                        log=True,
                                        sequence=50,
                                        protocol="icmp",
                                        protocol_options=dict(
                                            icmp=dict(
                                                administratively_prohibited=True
                                            )
                                        ),
                                    ),
                                    dict(remark="Overridden ACL"),
                                ],
                            )
                        ],
                    )
                ],
                state="overridden",
            )
        )
        commands = [
            "no ip access-list ACL1v4",
            "no ipv6 access-list ACL1v6",
            "ip access-list ACL2v4",
            "50 permit icmp any host 192.0.2.28 administratively-prohibited log",
            "remark Overridden ACL",
        ]
        self.execute_module(changed=True, commands=commands)

    def test_nxos_acls_overridden_idempotent(self):
        set_module_args(
            dict(
                config=[
                    dict(
                        afi="ipv4",
                        acls=[
                            dict(
                                name="ACL1v4",
                                aces=[
                                    dict(
                                        grant="permit",
                                        destination=dict(any=True),
                                        source=dict(any=True),
                                        sequence=10,
                                        protocol="ip",
                                    ),
                                    dict(
                                        grant="deny",
                                        destination=dict(any=True),
                                        source=dict(any=True),
                                        sequence=20,
                                        protocol="udp",
                                    ),
                                ],
                            )
                        ],
                    ),
                    dict(
                        afi="ipv6",
                        acls=[
                            dict(
                                name="ACL1v6",
                                aces=[
                                    dict(
                                        grant="permit",
                                        destination=dict(any=True),
                                        source=dict(any=True),
                                        sequence=10,
                                        protocol="sctp",
                                    )
                                ],
                            )
                        ],
                    ),
                ],
                state="overridden",
            )
        )
        self.execute_module(changed=False, commands=[])

    def test_nxos_acls_deletedafi(self):
        set_module_args(dict(config=[dict(afi="ipv4")], state="deleted"))
        commands = ["no ip access-list ACL1v4"]
        self.execute_module(changed=True, commands=commands)

    def test_nxos_acls_deletedall(self):
        set_module_args(dict(config=[], state="deleted"))
        commands = ["no ipv6 access-list ACL1v6", "no ip access-list ACL1v4"]
        self.execute_module(changed=True, commands=commands)

    def test_nxos_acls_rendered(self):
        set_module_args(
            dict(
                config=[
                    dict(
                        afi="ipv4",
                        acls=[
                            dict(
                                name="ACL1v4",
                                aces=[
                                    dict(
                                        grant="permit",
                                        destination=dict(any=True),
                                        source=dict(any=True),
                                        sequence=10,
                                        protocol="ip",
                                    ),
                                    dict(
                                        grant="deny",
                                        destination=dict(any=True),
                                        source=dict(any=True),
                                        sequence=20,
                                        protocol="udp",
                                    ),
                                ],
                            )
                        ],
                    ),
                    dict(
                        afi="ipv6",
                        acls=[
                            dict(
                                name="ACL1v6",
                                aces=[
                                    dict(
                                        grant="permit",
                                        destination=dict(any=True),
                                        source=dict(any=True),
                                        sequence=10,
                                        protocol="sctp",
                                    )
                                ],
                            )
                        ],
                    ),
                ],
                state="rendered",
            )
        )
        commands = [
            "ip access-list ACL1v4",
            "10 permit ip any any",
            "20 deny udp any any",
            "ipv6 access-list ACL1v6",
            "10 permit sctp any any",
        ]
        result = self.execute_module(changed=False)
        self.assertEqual(
            sorted(result["rendered"]), sorted(commands), result["rendered"]
        )

    def test_nxos_acls_parsed(self):
        set_module_args(
            dict(
                running_config="""\nip access-list ACL1v4\n 10 permit ip any any\n 20 deny udp any any dscp AF23 precedence critical""",
                state="parsed",
            )
        )
        result = self.execute_module(changed=False)
        compare_list = [
            {
                "afi": "ipv4",
                "acls": [
                    {
                        "name": "ACL1v4",
                        "aces": [
                            {
                                "grant": "permit",
                                "sequence": 10,
                                "protocol": "ip",
                                "source": {"any": True},
                                "destination": {"any": True},
                            },
                            {
                                "grant": "deny",
                                "sequence": 20,
                                "protocol": "udp",
                                "source": {"any": True},
                                "destination": {"any": True},
                                "dscp": "AF23",
                                "precedence": "critical",
                            },
                        ],
                    }
                ],
            }
        ]
        self.assertEqual(result["parsed"], compare_list, result["parsed"])

    def test_nxos_acls_gathered(self):
        set_module_args(dict(config=[], state="gathered"))
        result = self.execute_module(changed=False)
        compare_list = [
            {
                "acls": [
                    {
                        "aces": [
                            {
                                "destination": {"any": True},
                                "sequence": 10,
                                "protocol": "sctp",
                                "source": {"any": True},
                                "grant": "permit",
                            }
                        ],
                        "name": "ACL1v6",
                    }
                ],
                "afi": "ipv6",
            },
            {
                "acls": [
                    {
                        "aces": [
                            {
                                "destination": {"any": True},
                                "sequence": 10,
                                "protocol": "ip",
                                "source": {"any": True},
                                "grant": "permit",
                            },
                            {
                                "destination": {"any": True},
                                "sequence": 20,
                                "protocol": "udp",
                                "source": {"any": True},
                                "grant": "deny",
                            },
                        ],
                        "name": "ACL1v4",
                    }
                ],
                "afi": "ipv4",
            },
        ]
        self.assertEqual(result["gathered"], compare_list, result["gathered"])
